from importlib.metadata import PackageNotFoundError, version

from .coords import *
from .facets import *
from .geoms import *
from .ggplot import ggplot, ggsave, save_as_pdf_pages
from .guides import *
from .labels import *
from .mapping import *
from .positions import *
from .qplot import qplot
from .scales import *
from .stats import *
from .themes import *
from .watermark import watermark

try:
    __version__ = version("plotnine")
except PackageNotFoundError:
    # package is not installed
    pass
finally:
    del version
    del PackageNotFoundError


def _get_all_imports():
    """
    Return list of all the imports

    This prevents sub-modules (geoms, stats, utils, ...)
    from being imported into the user namespace by the
    following import statement

        from plotnine import *

    This is because `from Module import Something`
    leads to `Module` itself coming into the namespace!!
    """
    from types import ModuleType

    lst = [
        name
        for name, obj in globals().items()
        if not (name.startswith("_") or isinstance(obj, ModuleType))
    ]
    return lst


__all__ = _get_all_imports()



from plotnine import (
    aes,
    as_labeller,
    facet_grid,
    facet_wrap,
    geom_point,
    ggplot,
    labeller,
)
from plotnine.data import mtcars


def number_to_word(n):
    lst = [
        "zero",
        "one",
        "two",
        "three",
        "four",
        "five",
        "six",
        "seven",
        "eight",
        "nine",
    ]
    try:
        return lst[int(n)]
    except IndexError:
        return str(n)


labeller_cols_both = labeller(
    rows="label_value", cols="label_both", multi_line=False
)

labeller_towords = labeller(
    rows="label_both",
    cols="label_both",
    multi_line=False,
    am=number_to_word,
    gear=number_to_word,
)

g = ggplot(mtcars, aes(x="wt", y="mpg")) + geom_point()


def test_label_value():
    p = g + facet_wrap("~ gear", labeller="label_value")

    assert p == "label_value"


def test_label_both():
    p = g + facet_wrap("~ gear", labeller="label_both")

    assert p == "label_both"


def test_label_context():
    p = g + facet_wrap("~ gear", labeller="label_context")

    assert p == "label_context"


def test_label_context_wrap2vars():
    p = g + facet_wrap("~ gear + am", labeller="label_context")

    assert p == "label_context_wrap2vars"


def test_labeller_cols_both_wrap():
    p = g + facet_wrap("~ gear + am", labeller=labeller_cols_both)

    assert p == "labeller_cols_both_wrap"


def test_labeller_cols_both_grid():
    p = g + facet_grid("gear ~ am", labeller=labeller_cols_both)

    assert p == "labeller_cols_both_grid"


def test_labeller_towords():
    p = g + facet_grid("gear ~ am", labeller=labeller_towords)

    assert p == "labeller_towords"


def test_aslabeller_func_hashtag():
    func = as_labeller(lambda s: f"#{s}")
    p = g + facet_wrap("~ gear + am", labeller=func)

    assert p == "aslabeller_func_hashtagit"


def test_aslabeller_dict_0tag():
    func = as_labeller({"0": "<tag>0</tag>"})
    p = g + facet_wrap("~ gear + am", labeller=func)

    assert p == "aslabeller_dict_0tag"


def test_uneven_num_of_lines():
    @as_labeller
    def func(s):
        if s == "3":
            s = f"{s}\nline2\nline3\nline4"
        return s

    p = g + facet_wrap("~ gear + am", labeller=func)
    # NOTE: The allocated space for the strip is still not yet
    # enough for the text
    assert p == "uneven_num_of_lines"



from __future__ import annotations

import typing

import numpy as np
import pandas as pd

from ..doctools import document
from ..exceptions import PlotnineError
from ..utils import SIZE_FACTOR, to_rgba
from .geom import geom
from .geom_point import geom_point
from .geom_polygon import geom_polygon

if typing.TYPE_CHECKING:
    from typing import Any

    import numpy.typing as npt
    from shapely.geometry.polygon import LinearRing, Polygon

    from plotnine.iapi import panel_view
    from plotnine.typing import (
        Aes,
        Axes,
        Coord,
        DataLike,
        DrawingArea,
        Layer,
        PathPatch,
    )


@document
class geom_map(geom):
    """
    Draw map feature

    The map feature are drawn without any special projections.

    {usage}

    Parameters
    ----------
    {common_parameters}

    Notes
    -----
    This geom is best suited for plotting a shapefile read into
    geopandas dataframe. The dataframe should have a ``geometry``
    column.
    """

    DEFAULT_AES = {
        "alpha": 1,
        "color": "#111111",
        "fill": "#333333",
        "linetype": "solid",
        "shape": "o",
        "size": 0.5,
        "stroke": 0.5,
    }
    DEFAULT_PARAMS = {
        "stat": "identity",
        "position": "identity",
        "na_rm": False,
    }
    REQUIRED_AES = {"geometry"}

    def __init__(
        self,
        mapping: Aes | None = None,
        data: DataLike | None = None,
        **kwargs: Any,
    ):
        geom.__init__(self, mapping, data, **kwargs)
        # Almost all geodataframes loaded from shapefiles
        # have a geometry column.
        if "geometry" not in self.mapping:
            self.mapping["geometry"] = "geometry"

    def setup_data(self, data: pd.DataFrame) -> pd.DataFrame:
        if not len(data):
            return data

        # Remove any NULL geometries, and remember
        # All the non-Null shapes in a shapefile are required to be
        # of the same shape type.
        bool_idx = np.array([g is not None for g in data["geometry"]])
        if not np.all(bool_idx):
            data = data.loc[bool_idx]

        # Add polygon limits. Scale training uses them
        try:
            bounds = data["geometry"].bounds
        except AttributeError:
            # The geometry is not a GeoSeries
            # Bounds calculation is extracted from
            # geopandas.base.GeoPandasBase.bounds
            bounds = pd.DataFrame(
                np.array([x.bounds for x in data["geometry"]]),
                columns=["xmin", "ymin", "xmax", "ymax"],
                index=data.index,
            )
        else:
            bounds.rename(
                columns={
                    "minx": "xmin",
                    "maxx": "xmax",
                    "miny": "ymin",
                    "maxy": "ymax",
                },
                inplace=True,
            )

        data = pd.concat([data, bounds], axis=1)
        return data

    def draw_panel(
        self,
        data: pd.DataFrame,
        panel_params: panel_view,
        coord: Coord,
        ax: Axes,
        **params: Any,
    ):
        if not len(data):
            return

        data.loc[data["color"].isnull(), "color"] = "none"
        data.loc[data["fill"].isnull(), "fill"] = "none"
        data["fill"] = to_rgba(data["fill"], data["alpha"])

        geom_type = data.geometry.iloc[0].geom_type
        if geom_type in ("Polygon", "MultiPolygon"):
            from matplotlib.collections import PatchCollection

            data["size"] *= SIZE_FACTOR
            patches = [PolygonPatch(g) for g in data["geometry"]]
            coll = PatchCollection(
                patches,
                edgecolor=data["color"],
                facecolor=data["fill"],
                linestyle=data["linetype"],
                linewidth=data["size"],
                zorder=params["zorder"],
                rasterized=params["raster"],
            )
            ax.add_collection(coll)
        elif geom_type == "Point":
            # Extract point coordinates from shapely geom
            # and plot with geom_point
            arr = np.array([list(g.coords)[0] for g in data["geometry"]])
            data["x"] = arr[:, 0]
            data["y"] = arr[:, 1]
            for _, gdata in data.groupby("group"):
                gdata.reset_index(inplace=True, drop=True)
                gdata.is_copy = None
                geom_point.draw_group(gdata, panel_params, coord, ax, **params)
        elif geom_type == "MultiPoint":
            # Where n is the length of the dataframe (no. of multipoints),
            #       m is the number of all points in all multipoints
            #
            # - MultiPoint -> List of Points (tuples) (n -> m)
            # - Explode the list, to create a dataframe were each point
            #      is associated with the right aesthetics (n -> m)
            # - Create x & y columns from the points (m -> m)
            data["points"] = [
                [p.coords[0] for p in mp.geoms] for mp in data["geometry"]
            ]
            data = data.explode("points", ignore_index=True)
            data["x"] = [p[0] for p in data["points"]]
            data["y"] = [p[1] for p in data["points"]]
            geom_point.draw_group(data, panel_params, coord, ax, **params)
        elif geom_type in ("LineString", "MultiLineString"):
            from matplotlib.collections import LineCollection

            data["size"] *= SIZE_FACTOR
            data["color"] = to_rgba(data["color"], data["alpha"])
            segments = []
            for g in data["geometry"]:
                if g.geom_type == "LineString":
                    segments.append(g.coords)
                else:
                    segments.extend(_g.coords for _g in g.geoms)

            coll = LineCollection(
                segments,
                edgecolor=data["color"],
                linewidth=data["size"],
                linestyle=data["linetype"],
                zorder=params["zorder"],
                rasterized=params["raster"],
            )
            ax.add_collection(coll)
        else:
            raise TypeError(f"Could not plot geometry of type '{geom_type}'")

    @staticmethod
    def draw_legend(
        data: pd.Series[Any], da: DrawingArea, lyr: Layer
    ) -> DrawingArea:
        """
        Draw a rectangle in the box

        Parameters
        ----------
        data : Series
            Data Row
        da : DrawingArea
            Canvas
        lyr : layer
            Layer

        Returns
        -------
        out : DrawingArea
        """
        data["size"] = data["stroke"]
        del data["stroke"]
        return geom_polygon.draw_legend(data, da, lyr)


def PolygonPatch(
    obj: Polygon,
) -> PathPatch:
    """
    Return a Matplotlib patch from a Polygon/MultiPolygon Geometry

    Parameters
    ----------
    obj : shapley.geometry.Polygon | shapley.geometry.MultiPolygon
        A Polygon or MultiPolygon to create a patch for description

    Returns
    -------
    result : matplotlib.patches.PathPatch
        A patch representing the shapely geometry

    Notes
    -----
    This functionality was originally provided by the descartes package
    by Sean Gillies (BSD license, https://pypi.org/project/descartes)
    which is nolonger being maintained.
    """
    from matplotlib.patches import PathPatch
    from matplotlib.path import Path

    def cw_coords(ring: LinearRing) -> npt.NDArray[Any]:
        """
        Return Clockwise array coordinates

        Parameters
        ----------
        ring: shapely.geometry.polygon.LinearRing
            LinearRing

        Returns
        -------
        out: ndarray
            (n x 2) array of coordinate points.
        """
        if ring.is_ccw:
            return np.asarray(ring.coords)[:, :2][::-1]
        return np.asarray(ring.coords)[:, :2]

    def ccw_coords(ring: LinearRing) -> npt.NDArray[Any]:
        """
        Return Counter Clockwise array coordinates

        Parameters
        ----------
        ring: shapely.geometry.polygon.LinearRing
            LinearRing

        Returns
        -------
        out: ndarray
            (n x 2) array of coordinate points.
        """
        if ring.is_ccw:
            return np.asarray(ring.coords)[:, :2]
        return np.asarray(ring.coords)[:, :2][::-1]

    # The interiors are holes in the Polygon
    # MPL draws a hole if the vertex points are specified
    # in an opposite direction. So we use Clockwise for
    # the exterior/shell and Counter-Clockwise for any
    # interiors/holes
    if obj.geom_type == "Polygon":
        _exterior = [Path(cw_coords(obj.exterior))]
        _interior = [Path(ccw_coords(ring)) for ring in obj.interiors]
    else:
        # A MultiPolygon has one or more Polygon geoms.
        # Concatenate the exterior of all the Polygons
        # and the interiors
        _exterior = []
        _interior = []
        for p in obj.geoms:
            _exterior.append(Path(cw_coords(p.exterior)))
            _interior.extend([Path(ccw_coords(ring)) for ring in p.interiors])

    path = Path.make_compound_path(*_exterior, *_interior)
    return PathPatch(path)


def check_geopandas():
    try:
        import geopandas  # noqa: F401
    except ImportError:
        raise PlotnineError(
            "geom_map requires geopandas. Please install geopandas."
        )



from __future__ import annotations

import typing
from itertools import cycle, islice

import numpy as np
import pandas as pd

from ..coords import coord_flip
from ..scales.scale import scale_discrete
from .annotate import annotate
from .geom import geom
from .geom_polygon import geom_polygon
from .geom_rect import geom_rect

if typing.TYPE_CHECKING:
    from typing import Any, Literal, Sequence

    from plotnine.iapi import panel_view
    from plotnine.typing import Axes, Coord, Layout, TupleFloat2


class annotation_stripes(annotate):
    """
    Alternating stripes, centered around each label.

    Useful as a background for geom_jitter.

    Parameters
    ----------
    fill : list-like
        List of colors for the strips. The default  is
        `("#AAAAAA", "#CCCCCC")`
    fill_range: 'cycle' | 'nocycle' | 'auto' | 'no'
        How to fill stripes beyond the range of scale::

            'cycle'      # keep cycling the colors of the
                         # stripes after the range ends
            'nocycle'    # stop cycling the colors of the
                         # stripes after the range ends
            'auto'       # 'cycle' for continuous scales and
                         # 'nocycle' for discrete scales.
            'no'         # Do not add stripes passed the range
                         # passed the range of the scales

        Default is 'auto'.
    direction : 'vertical' or 'horizontal'
        Orientation of the stripes
    extend : tuple
        Range of the stripes. The default is (0, 1), top to bottom.
        The values should be in the range [0, 1].
    **kwargs : dict
        Other aesthetic parameters for the rectangular stripes.
        They include; *alpha*, *color*, *linetype*, and *size*.
    """

    def __init__(
        self,
        fill: Sequence[str] = ("#AAAAAA", "#CCCCCC"),
        fill_range: Literal["auto", "cycle", "no", "nocycle"] = "auto",
        direction: Literal["horizontal", "vertical"] = "vertical",
        extend: TupleFloat2 = (0, 1),
        **kwargs: Any,
    ):
        allowed = ("vertical", "horizontal")
        if direction not in allowed:
            raise ValueError(f"direction must be one of {allowed}")
        self._annotation_geom = _geom_stripes(
            fill=fill,
            fill_range=fill_range,
            extend=extend,
            direction=direction,
            **kwargs,
        )


class _geom_stripes(geom):
    DEFAULT_AES = {}
    REQUIRED_AES = set()
    DEFAULT_PARAMS = {
        "stat": "identity",
        "position": "identity",
        "na_rm": False,
        "color": None,
        "fill": ("#AAAAAA", "#CCCCCC"),
        "linetype": "solid",
        "size": 1,
        "alpha": 0.5,
        "direction": "vertical",
        "extend": (0, 1),
        "fill_range": "auto",
    }
    draw_legend = staticmethod(geom_polygon.draw_legend)

    def draw_layer(
        self, data: pd.DataFrame, layout: Layout, coord: Coord, **params: Any
    ):
        """
        Draw stripes on every panel
        """
        for pid in layout.layout["PANEL"]:
            ploc = pid - 1
            panel_params = layout.panel_params[ploc]
            ax = layout.axs[ploc]
            self.draw_group(data, panel_params, coord, ax, **params)

    @staticmethod
    def draw_group(
        data: pd.DataFrame,
        panel_params: panel_view,
        coord: Coord,
        ax: Axes,
        **params: Any,
    ):
        extend = params["extend"]
        fill_range = params["fill_range"]
        direction = params["direction"]

        # Range
        if direction == "vertical":
            axis, other_axis = "x", "y"
        else:
            axis, other_axis = "y", "x"

        if isinstance(coord, coord_flip):
            axis, other_axis = other_axis, axis

        _axis = getattr(panel_params, axis)
        breaks = _axis.breaks
        range = _axis.range
        other_range = getattr(panel_params, other_axis).range

        if fill_range == "auto":
            if isinstance(_axis.scale, scale_discrete):
                fill_range = "nocycle"
            else:
                fill_range = "cycle"

        # Breaks along the width
        n_stripes = len(breaks)
        if n_stripes > 1:
            diff = np.diff(breaks)
            step = diff[0]
            equal_spaces = np.all(diff == step)
            if not equal_spaces:
                raise ValueError(
                    "The major breaks are not equally spaced. "
                    "We cannot create stripes."
                )
        else:
            step = breaks[0]

        deltas = np.array([step / 2] * n_stripes)
        many_stripes = len(breaks) > 1
        xmin = breaks - deltas
        xmax = breaks + deltas
        if fill_range in ("cycle", "nocycle") and many_stripes:
            if range[0] < breaks[0]:
                n_stripes += 1
                xmax = np.insert(xmax, 0, xmin[0])
                xmin = np.insert(xmin, 0, range[0])
            if range[1] > breaks[1]:
                n_stripes += 1
                xmin = np.append(xmin, xmax[-1])
                xmax = np.append(xmax, range[1])

        # Height
        full_height = other_range[1] - other_range[0]
        ymin = other_range[0] + full_height * extend[0]
        ymax = other_range[0] + full_height * extend[1]
        fill = list(islice(cycle(params["fill"]), n_stripes))
        if fill_range == "nocycle" and many_stripes:
            # there are at least two stripes at this point
            fill[0] = fill[1]
            fill[-1] = fill[-2]

        if direction != "vertical":
            xmin, xmax, ymin, ymax = ymin, ymax, xmin, xmax

        data = pd.DataFrame(
            {
                "xmin": xmin,
                "xmax": xmax,
                "ymin": ymin,
                "ymax": ymax,
                "fill": fill,
                "alpha": params["alpha"],
                "color": params["color"],
                "linetype": params["linetype"],
                "size": params["size"],
            }
        )

        return geom_rect.draw_group(data, panel_params, coord, ax, **params)



from __future__ import annotations

import typing
from copy import deepcopy

import pandas as pd

from ..exceptions import PlotnineError
from ..layer import layer
from ..mapping import aes
from ..utils import (
    Registry,
    check_required_aesthetics,
    copy_keys,
    data_mapping_as_kwargs,
    groupby_apply,
    is_string,
    remove_missing,
    uniquecols,
)

if typing.TYPE_CHECKING:
    from typing import Any

    from plotnine.iapi import pos_scales
    from plotnine.typing import (
        Aes,
        DataLike,
        EvalEnvironment,
        Geom,
        Layout,
    )


class stat(metaclass=Registry):
    """Base class of all stats"""

    __base__ = True

    REQUIRED_AES = set()
    DEFAULT_AES = {}
    NON_MISSING_AES = set()
    DEFAULT_PARAMS = {}

    # Should the values produced by the statistic also
    # be transformed in the second pass when recently
    # added statistics are trained to the scales
    retransform = True

    # Stats may modify existing columns or create extra
    # columns.
    #
    # Any extra columns that may be created by the stat
    # should be specified in this set
    # see: stat_bin
    CREATES = set()

    # Documentation for the aesthetics. It ie added under the
    # documentation for mapping parameter. Use {aesthetics_table}
    # placeholder to insert a table for all the aesthetics and
    # their default values.
    _aesthetics_doc = "{aesthetics_table}"

    # Plot namespace, it gets its value when the plot is being
    # built.
    environment: EvalEnvironment | None = None

    def __init__(
        self,
        mapping: Aes | None = None,
        data: DataLike | None = None,
        **kwargs: Any,
    ):
        kwargs = data_mapping_as_kwargs((data, mapping), kwargs)
        self._kwargs = kwargs  # Will be used to create the geom
        self.params = copy_keys(kwargs, deepcopy(self.DEFAULT_PARAMS))
        self.DEFAULT_AES = aes(**self.DEFAULT_AES)
        self.aes_params = {
            ae: kwargs[ae] for ae in (self.aesthetics() & kwargs.keys())
        }

    @staticmethod
    def from_geom(geom: Geom) -> stat:
        """
        Return an instantiated stat object

        stats should not override this method.

        Parameters
        ----------
        geom : geom
            `geom`

        Returns
        -------
        out : stat
            A stat object

        Raises
        ------
        :class:`PlotnineError` if unable to create a `stat`.
        """
        name = geom.params["stat"]
        kwargs = geom._kwargs
        # More stable when reloading modules than
        # using issubclass
        if not isinstance(name, type) and hasattr(name, "compute_layer"):
            return name

        if isinstance(name, stat):
            return name
        elif isinstance(name, type) and issubclass(name, stat):
            klass = name
        elif is_string(name):
            if not name.startswith("stat_"):
                name = f"stat_{name}"
            klass = Registry[name]
        else:
            raise PlotnineError(f"Unknown stat of type {type(name)}")

        valid_kwargs = (
            klass.aesthetics() | klass.DEFAULT_PARAMS.keys()
        ) & kwargs.keys()

        params = {k: kwargs[k] for k in valid_kwargs}
        return klass(geom=geom, **params)

    def __deepcopy__(self, memo: dict[Any, Any]) -> stat:
        """
        Deep copy without copying the self.data dataframe

        stats should not override this method.
        """
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        old = self.__dict__
        new = result.__dict__

        # don't make a _kwargs and environment
        shallow = {"_kwargs", "environment"}
        for key, item in old.items():
            if key in shallow:
                new[key] = old[key]
                memo[id(new[key])] = new[key]
            else:
                new[key] = deepcopy(old[key], memo)

        return result

    @classmethod
    def aesthetics(cls) -> set[str]:
        """
        Return a set of all non-computed aesthetics for this stat.

        stats should not override this method.
        """
        aesthetics = cls.REQUIRED_AES.copy()
        calculated = aes(**cls.DEFAULT_AES)._calculated
        for ae in set(cls.DEFAULT_AES) - set(calculated):
            aesthetics.add(ae)
        return aesthetics

    def use_defaults(self, data: pd.DataFrame) -> pd.DataFrame:
        """
        Combine data with defaults and set aesthetics from parameters

        stats should not override this method.

        Parameters
        ----------
        data : dataframe
            Data used for drawing the geom.

        Returns
        -------
        out : dataframe
            Data used for drawing the geom.
        """
        missing = (
            self.aesthetics() - set(self.aes_params.keys()) - set(data.columns)
        )

        for ae in missing - self.REQUIRED_AES:
            if self.DEFAULT_AES[ae] is not None:
                data[ae] = self.DEFAULT_AES[ae]

        missing = self.aes_params.keys() - set(data.columns)

        for ae in self.aes_params:
            data[ae] = self.aes_params[ae]

        return data

    def setup_params(self, data: pd.DataFrame) -> dict[str, Any]:
        """
        Overide this to verify or adjust parameters

        Parameters
        ----------
        data : dataframe
            Data

        Returns
        -------
        out : dict
            Parameters used by the stats.
        """
        return self.params

    def setup_data(self, data: pd.DataFrame) -> pd.DataFrame:
        """
        Overide to modify data before compute_layer is called

        Parameters
        ----------
        data : dataframe
            Data

        Returns
        -------
        out : dataframe
            Data
        """
        return data

    def finish_layer(
        self, data: pd.DataFrame, params: dict[str, Any]
    ) -> pd.DataFrame:
        """
        Modify data after the aesthetics have been mapped

        This can be used by stats that require access to the mapped
        values of the computed aesthetics, part 3 as shown below.

            1. stat computes and creates variables
            2. variables mapped to aesthetics
            3. stat sees and modifies data according to the
               aesthetic values

        The default to is to do nothing.

        Parameters
        ----------
        data : dataframe
            Data for the layer
        params : dict
            Paremeters

        Returns
        -------
        data : dataframe
            Modified data
        """
        return data

    @classmethod
    def compute_layer(
        cls, data: pd.DataFrame, params: dict[str, Any], layout: Layout
    ) -> pd.DataFrame:
        """
        Calculate statistics for this layers

        This is the top-most computation method for the
        stat. It does not do any computations, but it
        knows how to verify the data, partition it call the
        next computation method and merge results.

        stats should not override this method.

        Parameters
        ----------
        data : panda.DataFrame
            Data points for all objects in a layer.
        params : dict
            Stat parameters
        layout : plotnine.layout.Layout
            Panel layout information
        """
        check_required_aesthetics(
            cls.REQUIRED_AES,
            list(data.columns) + list(params.keys()),
            cls.__name__,
        )

        data = remove_missing(
            data,
            na_rm=params.get("na_rm", False),
            vars=list(cls.REQUIRED_AES | cls.NON_MISSING_AES),
            name=cls.__name__,
            finite=True,
        )

        def fn(pdata):
            """
            Compute function helper
            """
            # Given data belonging to a specific panel, grab
            # the corresponding scales and call the method
            # that does the real computation
            if len(pdata) == 0:
                return pdata
            pscales = layout.get_scales(pdata["PANEL"].iat[0])
            return cls.compute_panel(pdata, pscales, **params)

        return groupby_apply(data, "PANEL", fn)

    @classmethod
    def compute_panel(
        cls, data: pd.DataFrame, scales: pos_scales, **params: Any
    ):
        """
        Calculate the statistics for all the groups

        Return the results in a single dataframe.

        This is a default function that can be overriden
        by individual stats

        Parameters
        ----------
        data : dataframe
            data for the computing
        scales : dataclass
            x (``scales.x``) and y (``scales.y``) scale objects.
            The most likely reason to use scale information is
            to find out the physical size of a scale. e.g::

                range_x = scales.x.dimension()

        params : dict
            The parameters for the stat. It includes default
            values if user did not set a particular parameter.
        """
        if not len(data):
            return type(data)()

        stats = []
        for _, old in data.groupby("group"):
            new = cls.compute_group(old, scales, **params)
            unique = uniquecols(old)
            missing = unique.columns.difference(new.columns)
            idx = [0] * len(new)
            u = unique.loc[idx, missing].reset_index(  # pyright: ignore
                drop=True
            )
            # concat can have problems with empty dataframes that
            # have an index
            if u.empty and len(u):
                u = type(data)()

            df = pd.concat([new, u], axis=1)
            stats.append(df)

        stats = pd.concat(stats, axis=0, ignore_index=True)
        # Note: If the data coming in has columns with non-unique
        # values with-in group(s), this implementation loses the
        # columns. Individual stats may want to do some preparation
        # before then fall back on this implementation or override
        # it completely.
        return stats

    @classmethod
    def compute_group(
        cls, data: pd.DataFrame, scales: pos_scales, **params: Any
    ) -> pd.DataFrame:
        """
        Calculate statistics for the group

        All stats should implement this method

        Parameters
        ----------
        data : dataframe
            Data for a group
        scales : types.SimpleNamespace
            x (``scales.x``) and y (``scales.y``) scale objects.
            The most likely reason to use scale information is
            to find out the physical size of a scale. e.g::

                range_x = scales.x.dimension()

        params : dict
            Parameters
        """
        msg = "{} should implement this method."
        raise NotImplementedError(msg.format(cls.__name__))

    def __radd__(self, gg):
        """
        Add layer representing stat object on the right

        Parameters
        ----------
        gg : ggplot
            ggplot object

        Returns
        -------
        out : ggplot
            ggplot object with added layer
        """
        gg += self.to_layer()  # Add layer
        return gg

    def to_layer(self):
        """
        Make a layer that represents this stat

        Returns
        -------
        out : layer
            Layer
        """
        # Create, geom from stat, then layer from geom
        from ..geoms.geom import geom

        return layer.from_geom(geom.from_stat(self))



import numpy as np
import pandas as pd

from plotnine import (
    aes,
    coord_trans,
    geom_point,
    geom_rect,
    geom_tile,
    ggplot,
    labs,
)

n = 4

df = pd.DataFrame(
    {
        "xmin": range(1, n * 2, 2),
        "xmax": range(2, n * 2 + 1, 2),
        "ymin": [1] * n,
        "ymax": [2] * n,
        "z": range(n),
    }
)

# geom_rect and geom_tile are similar but
# parameterised different
# for geom_tile
df["x"] = df["xmin"] + 0.5
df["y"] = df["ymin"] + 0.5


def test_rect_aesthetics():
    p = (
        ggplot(df, aes(xmin="xmin", xmax="xmax", ymin="ymin", ymax="ymax"))
        + geom_rect()
        + geom_rect(
            aes(ymin="ymin+2", ymax="ymax+2", alpha="z"), show_legend=False
        )
        + geom_rect(aes(ymin="ymin+4", ymax="ymax+4", fill="factor(z)"))
        + geom_rect(
            aes(ymin="ymin+6", ymax="ymax+6", color="factor(z+1)"), size=2
        )
        + geom_rect(
            aes(ymin="ymin+8", ymax="ymax+8", linetype="factor(z+2)"),
            color="yellow",
            size=2,
        )
        +
        # for comparison with geom_tile which
        # has labels by default
        labs(x="x", y="y")
    )

    assert p == "rect-aesthetics"


def test_rect_nofill():
    p = (
        ggplot(df)
        + aes(xmin="xmin", xmax="xmax", ymin="ymin", ymax="ymax")
        + geom_rect(color="red", fill=None, size=2)
        + geom_rect(
            aes(ymin="ymin+2", ymax="ymax+2"),
            color="blue",
            fill="None",
            size=2,
        )
        + geom_rect(
            aes(ymin="ymin+4", ymax="ymax+4"), color="green", fill="", size=2
        )
        + geom_rect(
            aes(ymin="ymin+6", ymax="ymax+6"),
            color="yellow",
            fill="gray",
            size=2,
        )
    )

    assert p == "rect-nofill"


def test_tile_aesthetics():
    p = (
        ggplot(df, aes("x", "y", width=1, height=1))
        + geom_tile()
        + geom_tile(aes(y="y+2", alpha="z"), show_legend=False)
        + geom_tile(aes(y="y+4", fill="factor(z)"))
        + geom_tile(aes(y="y+6", color="factor(z+1)"), size=2)
        + geom_tile(
            aes(y="y+8", linetype="factor(z+2)"), color="yellow", size=2
        )
    )

    assert p == "tile-aesthetics"


def test_infinite_rects():
    df = pd.DataFrame({"x": range(10), "y": range(10)})
    rdf = pd.DataFrame(
        {
            "xmin": [3],
            "xmax": 7,
            "ymin": -np.inf,
            "ymax": np.inf,
        }
    )

    p = (
        ggplot(df, aes("x", "y"))
        + geom_point()
        + geom_rect(
            data=rdf,
            mapping=aes(xmin="xmin", xmax="xmax", ymin="ymin", ymax="ymax"),
            alpha=0.2,
            inherit_aes=False,
        )
    )

    assert p == "infinite-rects"


def test_coord_trans():
    df = pd.DataFrame({"x": range(10), "y": range(10)})
    rdf = pd.DataFrame(
        {
            "xmin": [3],
            "xmax": 7,
            "ymin": -np.inf,
            "ymax": np.inf,
        }
    )

    p = (
        ggplot(df, aes("x", "y"))
        + geom_point()
        + geom_rect(
            data=rdf,
            mapping=aes(xmin="xmin", xmax="xmax", ymin="ymin", ymax="ymax"),
            alpha=0.2,
            inherit_aes=False,
        )
        + coord_trans()
    )

    assert p == "coord-trans"


def test_coord_trans_groups():
    df = pd.DataFrame(
        {
            "xmin": [0, 2, 4],
            "xmax": [1, 3, 5],
            "ymin": [0, 2, 4],
            "ymax": [1, 3, 5],
            "c": list("abc"),
        }
    )

    p = (
        ggplot(df)
        + geom_rect(
            aes(xmin="xmin", xmax="xmax", ymin="ymin", ymax="ymax", fill="c")
        )
        + coord_trans()
    )
    assert p == "coord-trans-groups"



import numpy as np
import pandas as pd

from ..doctools import document
from ..exceptions import PlotnineError
from .stat import stat
from .stat_qq import stat_qq


@document
class stat_qq_line(stat):
    """
    Calculate line through quantile-quantile plot

    {usage}

    Parameters
    ----------
    {common_parameters}
    distribution : str (default: norm)
        Distribution or distribution function name. The default is
        *norm* for a normal probability plot. Objects that look enough
        like a stats.distributions instance (i.e. they have a ppf
        method) are also accepted. See :mod:`scipy stats <scipy.stats>`
        for available distributions.
    dparams : dict, optional
        Distribution-specific shape parameters (shape parameters plus
        location and scale).
    quantiles : array_like, optional
        Probability points at which to calculate the theoretical
        quantile values. If provided, must be the same number as
        as the sample data points. The default is to use calculated
        theoretical points, use to ``alpha_beta`` control how
        these points are generated.
    alpha_beta : tuple
        Parameter values to use when calculating the quantiles.
        Default is :py:`(3/8, 3/8)`.
    line_p : tuple, optional
        Quantiles to use when fitting a Q-Q line. Must be 2 values.
        Default is :py:`(0.25, 0.75)`.
    fullrange : bool
        If :py:`True` the fit will span the full range of the plot.

    See Also
    --------
    scipy.stats.mstats.plotting_positions : Uses ``alpha_beta``
        to calculate the quantiles.
    """

    REQUIRED_AES = {"sample"}
    DEFAULT_PARAMS = {
        "geom": "qq_line",
        "position": "identity",
        "na_rm": False,
        "distribution": "norm",
        "dparams": {},
        "quantiles": None,
        "alpha_beta": (3 / 8, 3 / 8),
        "line_p": (0.25, 0.75),
        "fullrange": False,
    }
    CREATES = {"x", "y"}

    def setup_params(self, data):
        if len(self.params["line_p"]) != 2:
            raise PlotnineError(
                "Cannot fit line quantiles. " "'line_p' must be of length 2"
            )
        return self.params

    @classmethod
    def compute_group(cls, data, scales, **params):
        from scipy.stats.mstats import mquantiles

        from .distributions import get_continuous_distribution

        line_p = params["line_p"]
        dparams = params["dparams"]

        # Compute theoretical values
        df = stat_qq.compute_group(data, scales, **params)
        sample = df["sample"].values
        theoretical = df["theoretical"].to_numpy()

        # Compute slope & intercept of the line through the quantiles
        cdist = get_continuous_distribution(params["distribution"])
        x_coords = cdist.ppf(line_p, **dparams)
        y_coords = mquantiles(sample, line_p)
        slope = (np.diff(y_coords) / np.diff(x_coords))[0]
        intercept = y_coords[0] - slope * x_coords[0]

        # Get x,y points that describe the line
        if params["fullrange"] and scales.x:
            x = scales.x.dimension()
        else:
            x = theoretical.min(), theoretical.max()

        x = np.asarray(x)
        y = slope * x + intercept
        data = pd.DataFrame({"x": x, "y": y})
        return data



from warnings import warn

import pandas as pd

from ..doctools import document
from ..exceptions import PlotnineWarning
from .stat import stat


# method_args are any of the keyword args (other than q) for
# statsmodels.regression.quantile_regression.QuantReg.fit
@document
class stat_quantile(stat):
    """
    Compute quantile regression lines

    {usage}

    Parameters
    ----------
    {common_parameters}
    quantiles : tuple, optional (default: (0.25, 0.5, 0.75))
        Quantiles of y to compute
    formula : str, optional (default: 'y ~ x')
        Formula relating y variables to x variables
    method_args : dict, optional
        Extra arguments passed on to the model fitting method,
        :meth:`statsmodels.regression.quantile_regression.QuantReg.fit`.

    See Also
    --------
    statsmodels.regression.quantile_regression.QuantReg
    plotnine.geoms.geom_quantile
    """

    _aesthetics_doc = """
    {aesthetics_table}

    .. rubric:: Options for computed aesthetics

    ::

         'quantile'  # quantile
         'group'     # group identifier

    Calculated aesthetics are accessed using the `after_stat` function.
    e.g. :py:`after_stat('quantile')`.
    """

    REQUIRED_AES = {"x", "y"}
    DEFAULT_PARAMS = {
        "geom": "quantile",
        "position": "identity",
        "na_rm": False,
        "quantiles": (0.25, 0.5, 0.75),
        "formula": "y ~ x",
        "method_args": {},
    }
    CREATES = {"quantile", "group"}

    def setup_params(self, data):
        params = self.params.copy()
        if params["formula"] is None:
            params["formula"] = "y ~ x"
            warn("Formula not specified, using '{}'", PlotnineWarning)
        try:
            iter(params["quantiles"])
        except TypeError:
            params["quantiles"] = (params["quantiles"],)

        return params

    @classmethod
    def compute_group(cls, data, scales, **params):
        res = [quant_pred(q, data, **params) for q in params["quantiles"]]
        return pd.concat(res, axis=0, ignore_index=True)


def quant_pred(q, data, **params):
    """
    Quantile precitions
    """
    import statsmodels.formula.api as smf

    mod = smf.quantreg(params["formula"], data)
    reg_res = mod.fit(q=q, **params["method_args"])
    out = pd.DataFrame(
        {
            "x": [data["x"].min(), data["x"].max()],
            "quantile": q,
            "group": "{}-{}".format(data["group"].iloc[0], q),
        }
    )
    out["y"] = reg_res.predict(out)
    return out



from contextlib import suppress
from warnings import warn

import numpy as np
import pandas as pd

from ..doctools import document
from ..exceptions import PlotnineError, PlotnineWarning
from ..mapping.evaluation import after_stat
from .stat import stat


# NOTE: Parameter discriptions are in
# statsmodels/nonparametric/kde.py
@document
class stat_density(stat):
    """
    Compute density estimate

    {usage}

    Parameters
    ----------
    {common_parameters}
    kernel : str, optional (default: 'gaussian')
        Kernel used for density estimation. One of::

            'biweight'
            'cosine'
            'cosine2'
            'epanechnikov'
            'gaussian'
            'triangular'
            'triweight'
            'uniform'

    adjust : float, optional (default: 1)
        An adjustment factor for the ``bw``. Bandwidth becomes
        :py:`bw * adjust`.
        Adjustment of the bandwidth.
    trim : bool, optional (default: False)
        This parameter only matters if you are displaying multiple
        densities in one plot. If :py:`False`, the default, each
        density is computed on the full range of the data. If
        :py:`True`, each density is computed over the range of that
        group; this typically means the estimated x values will not
        line-up, and hence you won't be able to stack density values.
    n : int, optional(default: 1024)
        Number of equally spaced points at which the density is to
        be estimated. For efficient computation, it should be a power
        of two.
    gridsize : int, optional (default: None)
        If gridsize is :py:`None`, :py:`max(len(x), 50)` is used.
    bw : str or float, optional (default: 'nrd0')
        The bandwidth to use, If a float is given, it is the bandwidth.
        The :py:`str` choices are::

            'nrd0'
            'normal_reference'
            'scott'
            'silverman'

        ``nrd0`` is a port of ``stats::bw.nrd0`` in R; it is eqiuvalent
        to ``silverman`` when there is more than 1 value in a group.
    cut : float, optional (default: 3)
        Defines the length of the grid past the lowest and highest
        values of ``x`` so that the kernel goes to zero. The end points
        are ``-/+ cut*bw*{min(x) or max(x)}``.
    clip : tuple, optional (default: (-np.inf, np.inf))
        Values in ``x`` that are outside of the range given by clip are
        dropped. The number of values in ``x`` is then shortened.

    See Also
    --------
    plotnine.geoms.geom_density
    statsmodels.nonparametric.kde.KDEUnivariate
    statsmodels.nonparametric.kde.KDEUnivariate.fit
    """

    _aesthetics_doc = """
    {aesthetics_table}

    .. rubric:: Options for computed aesthetics

    ::

        'density'   # density estimate

        'count'     # density * number of points,
                    # useful for stacked density plots

        'scaled'    # density estimate, scaled to maximum of 1

    """
    REQUIRED_AES = {"x"}
    DEFAULT_PARAMS = {
        "geom": "density",
        "position": "stack",
        "na_rm": False,
        "kernel": "gaussian",
        "adjust": 1,
        "trim": False,
        "n": 1024,
        "gridsize": None,
        "bw": "nrd0",
        "cut": 3,
        "clip": (-np.inf, np.inf),
    }
    DEFAULT_AES = {"y": after_stat("density")}
    CREATES = {"density", "count", "scaled", "n"}

    def setup_params(self, data):
        params = self.params.copy()
        lookup = {
            "biweight": "biw",
            "cosine": "cos",
            "cosine2": "cos2",
            "epanechnikov": "epa",
            "gaussian": "gau",
            "triangular": "tri",
            "triweight": "triw",
            "uniform": "uni",
        }

        with suppress(KeyError):
            params["kernel"] = lookup[params["kernel"].lower()]

        if params["kernel"] not in lookup.values():
            msg = (
                "kernel should be one of {}. "
                "You may use the abbreviations {}"
            )
            raise PlotnineError(msg.format(lookup.keys(), lookup.values()))

        return params

    @classmethod
    def compute_group(cls, data, scales, **params):
        weight = data.get("weight")

        if params["trim"]:
            range_x = data["x"].min(), data["x"].max()
        else:
            range_x = scales.x.dimension()

        return compute_density(data["x"], weight, range_x, **params)


def compute_density(x, weight, range, **params):
    """
    Compute density
    """
    import statsmodels.api as sm

    x = np.asarray(x, dtype=float)
    not_nan = ~np.isnan(x)
    x = x[not_nan]
    bw = params["bw"]
    kernel = params["kernel"]
    n = len(x)

    if n == 0 or (n == 1 and isinstance(bw, str)):
        if n == 1:
            warn(
                "To compute the density of a group with only one "
                "value set the bandwidth manually. e.g `bw=0.1`",
                PlotnineWarning,
            )
        warn(
            "Groups with fewer than 2 data points have been removed.",
            PlotnineWarning,
        )
        return pd.DataFrame()

    # kde is computed efficiently using fft. But the fft does
    # not support weights and is only available with the
    # gaussian kernel. When weights are relevant we
    # turn off the fft.
    if weight is None:
        if kernel != "gau":
            weight = np.ones(n) / n
    else:
        weight = np.asarray(weight, dtype=float)

    if kernel == "gau" and weight is None:
        fft = True
    else:
        fft = False

    if bw == "nrd0":
        bw = nrd0(x)

    kde = sm.nonparametric.KDEUnivariate(x)
    kde.fit(
        kernel=kernel,
        bw=bw,
        fft=fft,
        weights=weight,
        adjust=params["adjust"],
        cut=params["cut"],
        gridsize=params["gridsize"],
        clip=params["clip"],
    )

    x2 = np.linspace(range[0], range[1], params["n"])

    try:
        y = kde.evaluate(x2)
        if np.isscalar(y) and np.isnan(y):
            raise ValueError("kde.evaluate returned nan")
    except ValueError:
        y = []
        for _x in x2:
            result = kde.evaluate(_x)
            if isinstance(result, float):
                y.append(result)
            else:
                y.append(result[0])

    y = np.asarray(y)

    # Evaluations outside the kernel domain return np.nan,
    # these values and corresponding x2s are dropped.
    # The kernel domain is defined by the values in x, but
    # the evaluated values in x2 could have a much wider range.
    not_nan = ~np.isnan(y)
    x2 = x2[not_nan]
    y = y[not_nan]
    return pd.DataFrame(
        {
            "x": x2,
            "density": y,
            "scaled": y / np.max(y) if len(y) else [],
            "count": y * n,
            "n": n,
        }
    )


def nrd0(x):
    """
    Port of R stats::bw.nrd0

    This is equivalent to statsmodels silverman when x has more than
    1 unique value. It can never give a zero bandwidth.

    Parameters
    ----------
    x : array_like
        Values whose density is to be estimated

    Returns
    -------
    out : float
        Bandwidth of x
    """
    from scipy.stats import iqr

    n = len(x)
    if n < 1:
        raise ValueError(
            "Need at leat 2 data points to compute the nrd0 bandwidth."
        )

    std: float = np.std(x, ddof=1)  # pyright: ignore
    std_estimate: float = iqr(x) / 1.349
    low_std = np.min((std, std_estimate))
    if low_std == 0:
        low_std = std_estimate or np.abs(np.asarray(x)[0]) or 1
    return 0.9 * low_std * (n**-0.2)



import numpy as np
import pandas as pd
import pytest

from plotnine import aes, geom_density, ggplot, lims
from plotnine.exceptions import PlotnineWarning

n = 6  # Some even number greater than 2

# ladder: 0 1 times, 1 2 times, 2 3 times, ...
df = pd.DataFrame(
    {
        "x": np.repeat(range(n + 1), range(n + 1)),
        "z": np.repeat(range(n // 2), range(3, n * 2, 4)),
    }
)

p = ggplot(df, aes("x", fill="factor(z)"))


def test_gaussian():
    p1 = p + geom_density(kernel="gaussian", alpha=0.3)
    assert p1 == "gaussian"


def test_gaussian_weighted():
    p1 = p + geom_density(aes(weight="x"), kernel="gaussian", alpha=0.3)
    assert p1 == "gaussian_weighted"


def test_gaussian_trimmed():
    p2 = p + geom_density(kernel="gaussian", alpha=0.3, trim=True)
    assert p2 == "gaussian-trimmed"


def test_triangular():
    p3 = p + geom_density(
        kernel="triangular", bw="normal_reference", alpha=0.3
    )  # other
    assert p3 == "triangular"


def test_few_datapoints():
    df = pd.DataFrame({"x": [1, 2, 2, 3, 3, 3], "z": list("abbccc")})

    # Bandwidth not set
    p = ggplot(df, aes("x", color="z")) + geom_density() + lims(x=(-3, 9))
    with pytest.warns(PlotnineWarning) as record:
        p.draw_test()

    record = list(record)  # iterate more than 1 time
    assert any("e.g `bw=0.1`" in str(r.message) for r in record)
    assert any("Groups with fewer than 2" in str(r.message) for r in record)

    p = ggplot(df, aes("x", color="z")) + geom_density(bw=0.1) + lims(x=(0, 4))
    assert p == "few_datapoints"



import pandas as pd
import pytest

from plotnine import aes, geom_point, ggplot, stat_identity
from plotnine.exceptions import PlotnineError
from plotnine.geoms.geom import geom

df = pd.DataFrame({"col1": [1, 2, 3, 4], "col2": 2, "col3": list("abcd")})


def test_geom_basics():
    class geom_abc(geom):
        DEFAULT_AES = {"color": None}
        DEFAULT_PARAMS = {"stat": "identity", "position": "identity"}

    g = geom_abc(data=df)
    assert g.data is df

    g = geom_abc(df)
    assert g.data is df

    # geom data should not mess with the main data
    df_copy = df.copy()
    p = ggplot(df, aes("col", "mpg")) + geom_abc(df_copy)
    assert p.data is df
    assert p.layers[0].geom.data is df_copy

    g = geom_abc(aes(color="col1"))
    assert g.mapping["color"] == "col1"

    g = geom_abc(mapping=aes(color="col2"))
    assert g.mapping["color"] == "col2"

    # Multiple mappings
    with pytest.raises(TypeError):
        g = geom_abc(aes(color="col1"), aes(color="co1"))

    # setting, not mapping
    g = geom_abc(color="blue")
    assert g.aes_params["color"] == "blue"


def test_geom_with_invalid_argument():
    class geom_abc(geom):
        DEFAULT_AES = {"color": None}
        DEFAULT_PARAMS = {"stat": "identity", "position": "identity"}

    with pytest.raises(PlotnineError):
        geom_abc(do_the_impossible=True)


def test_geom_from_stat():
    stat = stat_identity(geom="point")
    assert isinstance(geom.from_stat(stat), geom_point)

    stat = stat_identity(geom="geom_point")
    assert isinstance(geom.from_stat(stat), geom_point)

    stat = stat_identity(geom=geom_point())
    assert isinstance(geom.from_stat(stat), geom_point)

    stat = stat_identity(geom=geom_point)
    assert isinstance(geom.from_stat(stat), geom_point)



from ..doctools import document
from .geom_bar import geom_bar


@document
class geom_histogram(geom_bar):
    """
    Histogram

    {usage}

    Parameters
    ----------
    {common_parameters}

    See Also
    --------
    plotnine.geoms.geom_bar
    """

    DEFAULT_PARAMS = {"stat": "bin", "position": "stack", "na_rm": False}



from ..doctools import document
from .geom_point import geom_point


@document
class geom_pointdensity(geom_point):
    """
    Scatterplot with density estimation at each point

    {usage}

    Parameters
    ----------
    {common_parameters}
    """

    DEFAULT_PARAMS = {
        "stat": "pointdensity",
        "position": "identity",
        "na_rm": False,
    }



from pathlib import Path

import numpy as np
import pandas as pd
import pytest

from plotnine import aes, geom_path, geom_point, ggplot
from plotnine.exceptions import PlotnineError, PlotnineWarning
from plotnine.layer import Layers, layer

df = pd.DataFrame({"x": range(10), "y": range(10)})
colors = ["red", "green", "blue"]

n = 5000
prg = np.random.RandomState(123)
df_large = pd.DataFrame(
    {"x": prg.uniform(1, 1000, size=n), "y": prg.uniform(1, 1000, size=n)}
)


def _get_colors(p):
    return [l.geom.aes_params["color"] for l in p.layers]


class TestLayers:
    # Give each geom in the layer a different color which we
    # can used to test the ordering.
    lyrs = Layers(
        [
            geom_point(color=colors[0]),
            geom_point(color=colors[1]),
            geom_point(color=colors[2]),
        ]
    )

    def test_addition(self):
        p = ggplot(df, aes("x", "y"))
        p1 = p + self.lyrs[0] + self.lyrs[1] + self.lyrs[2]
        assert _get_colors(p1) == colors

        p2 = p + self.lyrs
        assert _get_colors(p2) == colors

        # Real layers
        lyrs = Layers(layer.from_geom(obj) for obj in self.lyrs)
        p3 = p + lyrs
        assert _get_colors(p3) == colors

        p += self.lyrs
        assert _get_colors(p) == colors

        with pytest.raises(PlotnineError):
            geom_point() + layer.from_geom(geom_point())

        with pytest.raises(PlotnineError):
            geom_point() + self.lyrs

    def test_slicing(self):
        p = ggplot(df, aes("x", "y"))

        _lyrs = self.lyrs[1:]
        assert isinstance(_lyrs, Layers)

        p1 = p + self.lyrs[0] + self.lyrs[2]
        assert _get_colors(p1) == colors[::2]

        p2 = p + self.lyrs[::2]
        assert _get_colors(p2) == colors[::2]

        p3 = p + (self.lyrs[:1] + self.lyrs[2:])
        assert _get_colors(p3) == colors[::2]


def test_inserting_layers():
    class as_first_layer:
        def __init__(self, obj):
            self.obj = obj

        def __radd__(self, gg):
            gg.layers.insert(0, self.obj.to_layer())
            return gg

        def __rsub__(self, gg):
            return self.__radd__(gg)

    p = (
        ggplot(df, aes("x", "y"))
        + geom_point(size=4)
        + as_first_layer(geom_point(color="cyan", size=8))
        - as_first_layer(geom_point(color="red", size=12))
    )

    assert p == "inserting_layers"


def test_layer_with_nodata():
    # no data but good mappings
    p = ggplot() + geom_point(aes([1, 2], [1, 2]))
    p.build_test()

    # no data, and unresolvable mappings
    p = ggplot(aes("x", "y")) + geom_point()
    with pytest.raises(PlotnineError) as pe:
        p.draw_test()

    assert "Could not evaluate the 'x' mapping:" in pe.value.message


class TestRasterizing:
    p = ggplot(df_large, aes("x", "y"))

    def _assert_raster_smaller(self, p_no_raster, p_raster):
        # Plot and check that the file sizes are smaller when
        # rastering. Then delete the files.
        geom_name = p_raster.layers[0].geom.__class__.__name__
        fn1 = Path(f"{geom_name}-no-raster.pdf")
        fn2 = Path(f"{geom_name}-raster.pdf")

        try:
            with pytest.warns(PlotnineWarning):
                p_no_raster.save(fn1)
                p_raster.save(fn2)
            assert fn1.stat().st_size > fn2.stat().st_size
        finally:
            fn1.unlink(missing_ok=True)
            fn2.unlink(missing_ok=True)

    def test_geom_point(self):
        p1 = self.p + geom_point()
        p2 = self.p + geom_point(raster=True)
        self._assert_raster_smaller(p1, p2)

    def test_geom_path(self):
        p1 = self.p + geom_path()
        p2 = self.p + geom_path(raster=True)
        self._assert_raster_smaller(p1, p2)



from __future__ import annotations

import typing
from warnings import warn

from ..doctools import document
from ..exceptions import PlotnineWarning
from ..utils import alias
from .scale import scale_discrete

if typing.TYPE_CHECKING:
    from plotnine.typing import ScaleBreaksRaw


@document
class _scale_manual(scale_discrete):
    """
    Abstract class for manual scales

    Parameters
    ----------
    {superclass_parameters}
    """

    def __init__(self, values, **kwargs):
        # Match the values of the scale with the breaks (if given)
        if "breaks" in kwargs:
            from collections.abc import Sized

            breaks: ScaleBreaksRaw = kwargs["breaks"]
            if isinstance(breaks, Sized):
                if len(breaks) == len(values):
                    values = dict(zip(breaks, values))

        self._values = values
        scale_discrete.__init__(self, **kwargs)

    def palette(self, n):
        max_n = len(self._values)
        if n > max_n:
            msg = (
                f"The palette of {self.__class__.__name__} can return a "
                f"maximum of {max_n} values. {n} were requested from it."
            )
            warn(msg, PlotnineWarning)
        return self._values


@document
class scale_color_manual(_scale_manual):
    """
    Custom discrete color scale

    Parameters
    ----------
    values : array_like | dict
        Colors that make up the palette. The values will be matched with
        the ``limits`` of the scale or the ``breaks`` if provided.
        If it is a dict then it should map data values to colors.

    {superclass_parameters}
    """

    _aesthetics = ["color"]
    na_value = "#7F7F7F"


@document
class scale_fill_manual(_scale_manual):
    """
    Custom discrete fill scale

    Parameters
    ----------
    values : array_like | dict
        Colors that make up the palette. The values will be matched with
        the ``limits`` of the scale or the ``breaks`` if provided.
        If it is a dict then it should map data values to colors.
    {superclass_parameters}
    """

    _aesthetics = ["fill"]
    na_value = "#7F7F7F"


@document
class scale_shape_manual(_scale_manual):
    """
    Custom discrete shape scale

    Parameters
    ----------
    values : array_like | dict
        Shapes that make up the palette. See
        :mod:`matplotlib.markers.` for list of all possible
        shapes. The values will be matched with the ``limits``
        of the scale or the ``breaks`` if provided.
        If it is a dict then it should map data values to shapes.
    {superclass_parameters}

    See Also
    --------
    :mod:`matplotlib.markers`
    """

    _aesthetics = ["shape"]


@document
class scale_linetype_manual(_scale_manual):
    """
    Custom discrete linetype scale

    Parameters
    ----------
    values : list-like | dict
        Linetypes that make up the palette.
        Possible values of the list are:

            1. Strings like

            ::

                'solid'                # solid line
                'dashed'               # dashed line
                'dashdot'              # dash-dotted line
                'dotted'               # dotted line
                'None' or ' ' or ''    # draw nothing

            2. Tuples of the form (offset, (on, off, on, off, ....))
               e.g. (0, (1, 1)), (1, (2, 2)), (2, (5, 3, 1, 3))

        The values will be matched with the ``limits`` of the scale
        or the ``breaks`` if provided.
        If it is a dict then it should map data values to linetypes.
    {superclass_parameters}

    See Also
    --------
    :mod:`matplotlib.markers`
    """

    _aesthetics = ["linetype"]

    def map(self, x, limits=None):
        result = super().map(x, limits)
        # Ensure that custom linetypes are tuples, so that they can
        # be properly inserted and extracted from the dataframe
        if len(result) and hasattr(result[0], "__hash__"):
            result = [x if isinstance(x, str) else tuple(x) for x in result]
        return result


@document
class scale_alpha_manual(_scale_manual):
    """
    Custom discrete alpha scale

    Parameters
    ----------
    values : array_like | dict
        Alpha values (in the [0, 1] range) that make up
        the palette. The values will be matched with the
        ``limits`` of the scale or the ``breaks`` if provided.
        If it is a dict then it should map data values to alpha
        values.
    {superclass_parameters}
    """

    _aesthetics = ["alpha"]


@document
class scale_size_manual(_scale_manual):
    """
    Custom discrete size scale

    Parameters
    ----------
    values : array_like | dict
        Sizes that make up the palette. The values will be matched
        with the ``limits`` of the scale or the ``breaks`` if provided.
        If it is a dict then it should map data values to sizes.
    {superclass_parameters}
    """

    _aesthetics = ["size"]


# American to British spelling
alias("scale_colour_manual", scale_color_manual)



import matplotlib.transforms as mtransforms
from matplotlib.offsetbox import AuxTransformBox, DrawingArea


class ColoredDrawingArea(DrawingArea):
    """
    A Drawing Area with a background color
    """

    def __init__(
        self,
        width: float,
        height: float,
        xdescent=0.0,
        ydescent=0.0,
        clip=True,
        color="none",
    ):
        from matplotlib.patches import Rectangle

        super().__init__(width, height, xdescent, ydescent, clip=clip)

        self.patch = Rectangle(
            (0, 0),
            width=width,
            height=height,
            facecolor=color,
            edgecolor="None",
            linewidth=0,
            antialiased=False,
        )
        self.add_artist(self.patch)


# Fix AuxTransformBox, Adds a dpi_transform
# See https://github.com/matplotlib/matplotlib/pull/7344
class MyAuxTransformBox(AuxTransformBox):
    def __init__(self):
        aux_transform = mtransforms.IdentityTransform()
        AuxTransformBox.__init__(self, aux_transform)
        self.dpi_transform = mtransforms.Affine2D()

    def get_transform(self):
        """
        Return the :class:`~matplotlib.transforms.Transform` applied
        to the children
        """
        return (
            self.aux_transform
            + self.ref_offset_transform
            + self.dpi_transform
            + self.offset_transform
        )

    def draw(self, renderer):
        """
        Draw the children
        """
        dpi_cor = renderer.points_to_pixels(1.0)
        self.dpi_transform.clear()
        self.dpi_transform.scale(dpi_cor, dpi_cor)

        for c in self._children:
            c.draw(renderer)

        self.stale = False



import numpy as np
import pandas as pd

from ..doctools import document
from ..exceptions import PlotnineError
from ..mapping.evaluation import after_stat
from ..utils import resolution
from .stat import stat


@document
class stat_count(stat):
    """
    Counts the number of cases at each x position

    {usage}

    Parameters
    ----------
    {common_parameters}
    width : float, optional (default: None)
        Bar width. By default, set to 90% of the
        resolution of the data

    See Also
    --------
    plotnine.stats.stat_bin
    """

    _aesthetics_doc = """
    {aesthetics_table}

    .. rubric:: Options for computed aesthetics

    ::

        'count'  # Number of observations at a position
        'prop'   # Ratio of points in the panel at a position

    """

    REQUIRED_AES = {"x"}
    DEFAULT_PARAMS = {
        "geom": "histogram",
        "position": "stack",
        "na_rm": False,
        "width": None,
    }
    DEFAULT_AES = {"y": after_stat("count")}
    CREATES = {"count", "prop"}

    def setup_params(self, data):
        params = self.params.copy()
        if params["width"] is None:
            params["width"] = resolution(data["x"], False) * 0.9

        return params

    @classmethod
    def compute_group(cls, data, scales, **params):
        x = data["x"]
        if ("y" in data) or ("y" in params):
            msg = "stat_count() must not be used with a y aesthetic"
            raise PlotnineError(msg)

        weight = data.get("weight", [1] * len(x))  # pyright: ignore
        width = params["width"]
        df = pd.DataFrame({"weight": weight, "x": x})
        # weighted frequency count
        count = df.pivot_table("weight", index=["x"], aggfunc=np.sum)["weight"]
        x = count.index
        count = count.to_numpy()
        return pd.DataFrame(
            {
                "count": count,
                "prop": count / np.abs(count).sum(),
                "x": x,
                "width": width,
            }
        )



import pandas as pd

from plotnine import aes, arrow, geom_segment, ggplot

n = 4

# stepped horizontal line segments
df = pd.DataFrame(
    {
        "x": range(1, n + 1),
        "xend": range(2, n + 2),
        "y": range(n, 0, -1),
        "yend": range(n, 0, -1),
        "z": range(1, n + 1),
    }
)


def test_aesthetics():
    p = (
        ggplot(df, aes("x", "y", xend="xend", yend="yend"))
        + geom_segment(size=2)
        +
        # Positive slope segments
        geom_segment(aes(yend="yend+1", color="factor(z)"), size=2)
        + geom_segment(aes(yend="yend+2", linetype="factor(z)"), size=2)
        + geom_segment(aes(yend="yend+3", size="z"), show_legend=False)
        + geom_segment(
            aes(yend="yend+4", alpha="z"), size=2, show_legend=False
        )
    )

    assert p == "aesthetics"


def test_arrow():
    p = (
        ggplot(df, aes("x", "y", xend="xend", yend="yend"))
        + geom_segment(aes("x+2", xend="xend+2"), arrow=arrow(), size=2)
        + geom_segment(
            aes("x+4", xend="xend+4"), arrow=arrow(ends="first"), size=2
        )
        + geom_segment(
            aes("x+6", xend="xend+6"), arrow=arrow(ends="both"), size=2
        )
    )

    assert p == "arrow"



from __future__ import annotations

import typing

import numpy as np
import pandas as pd

from ..doctools import document
from ..utils import SIZE_FACTOR, to_rgba
from .geom import geom
from .geom_polygon import geom_polygon

if typing.TYPE_CHECKING:
    from typing import Any

    from plotnine.iapi import panel_view
    from plotnine.typing import Axes, Coord


@document
class geom_rect(geom):
    """
    Rectangles

    {usage}

    Parameters
    ----------
    {common_parameters}
    """

    DEFAULT_AES = {
        "color": None,
        "fill": "#595959",
        "linetype": "solid",
        "size": 0.5,
        "alpha": 1,
    }
    REQUIRED_AES = {"xmax", "xmin", "ymax", "ymin"}
    DEFAULT_PARAMS = {
        "stat": "identity",
        "position": "identity",
        "na_rm": False,
    }
    draw_legend = staticmethod(geom_polygon.draw_legend)

    def draw_panel(
        self,
        data: pd.DataFrame,
        panel_params: panel_view,
        coord: Coord,
        ax: Axes,
        **params: Any,
    ):
        """
        Plot all groups
        """
        if not coord.is_linear:
            data = _rectangles_to_polygons(data)
            for _, gdata in data.groupby("group"):
                gdata.reset_index(inplace=True, drop=True)
                geom_polygon.draw_group(
                    gdata, panel_params, coord, ax, **params
                )
        else:
            self.draw_group(data, panel_params, coord, ax, **params)

    @staticmethod
    def draw_group(
        data: pd.DataFrame,
        panel_params: panel_view,
        coord: Coord,
        ax: Axes,
        **params: Any,
    ):
        from matplotlib.collections import PolyCollection

        data = coord.transform(data, panel_params, munch=True)
        data["size"] *= SIZE_FACTOR

        limits = zip(data["xmin"], data["xmax"], data["ymin"], data["ymax"])

        verts = [[(l, b), (l, t), (r, t), (r, b)] for (l, r, b, t) in limits]

        fill = to_rgba(data["fill"], data["alpha"])
        color = data["color"]

        # prevent unnecessary borders
        if all(color.isnull()):
            color = "none"

        col = PolyCollection(
            verts,
            facecolors=fill,
            edgecolors=color,
            linestyles=data["linetype"],
            linewidths=data["size"],
            zorder=params["zorder"],
            rasterized=params["raster"],
        )
        ax.add_collection(col)


def _rectangles_to_polygons(df: pd.DataFrame) -> pd.DataFrame:
    """
    Convert rect data to polygons

    Paramters
    ---------
    df : dataframe
        Dataframe with *xmin*, *xmax*, *ymin* and *ymax* columns,
        plus others for aesthetics ...

    Returns
    -------
    data : dataframe
        Dataframe with *x* and *y* columns, plus others for
        aesthetics ...
    """
    n = len(df)

    # Helper indexing arrays
    xmin_idx = np.tile([True, True, False, False], n)
    xmax_idx = ~xmin_idx
    ymin_idx = np.tile([True, False, False, True], n)
    ymax_idx = ~ymin_idx

    # There are 2 x and 2 y values for each of xmin, xmax, ymin & ymax
    # The positions are as layed out in the indexing arrays
    # x and y values
    x = np.empty(n * 4)
    y = np.empty(n * 4)
    x[xmin_idx] = df["xmin"].repeat(2)
    x[xmax_idx] = df["xmax"].repeat(2)
    y[ymin_idx] = df["ymin"].repeat(2)
    y[ymax_idx] = df["ymax"].repeat(2)

    # Aesthetic columns and others
    other_cols = df.columns.difference(
        ["x", "y", "xmin", "xmax", "ymin", "ymax"]
    )
    d = {str(col): np.repeat(df[col].to_numpy(), 4) for col in other_cols}
    data = pd.DataFrame({"x": x, "y": y, **d})
    return data



import pandas as pd
import pytest

from plotnine import aes, annotate, geom_point, geom_rect, geom_segment, ggplot
from plotnine.exceptions import PlotnineError

n = 4
df = pd.DataFrame({"x": range(n), "y": range(n)})


def test_multiple_annotation_geoms():
    p = (
        ggplot(df, aes("x", "y"))
        + geom_point()
        + annotate("point", 0, 1, color="red", size=5)
        + annotate("text", 1, 2, label="Text", color="red", size=15, angle=45)
        + annotate(
            geom_rect,
            xmin=1.8,
            xmax=2.2,
            ymin=2.8,
            ymax=3.2,
            size=1,
            color="red",
            alpha=0.3,
        )
        + annotate(
            geom_segment, x=2.8, y=3.8, xend=3.2, yend=4.2, color="red", size=1
        )
    )
    assert p == "multiple_annotation_geoms"


def test_non_geom_raises():
    with pytest.raises(PlotnineError):
        annotate("doesnotexist", x=1)

    with pytest.raises(PlotnineError):
        annotate(5)

    class NotAGeom:
        pass

    with pytest.raises(PlotnineError):
        annotate(NotAGeom)

    with pytest.raises(PlotnineError):
        annotate(geom_point())



from __future__ import annotations

import typing

from ..doctools import document
from .geom import geom
from .geom_path import geom_path
from .geom_segment import geom_segment

if typing.TYPE_CHECKING:
    from typing import Any

    import pandas as pd

    from plotnine.iapi import panel_view
    from plotnine.typing import Axes, Coord


@document
class geom_linerange(geom):
    """
    Vertical interval represented by lines

    {usage}

    Parameters
    ----------
    {common_parameters}
    """

    DEFAULT_AES = {
        "alpha": 1,
        "color": "black",
        "linetype": "solid",
        "size": 0.5,
    }
    REQUIRED_AES = {"x", "ymin", "ymax"}
    DEFAULT_PARAMS = {
        "stat": "identity",
        "position": "identity",
        "na_rm": False,
    }
    draw_legend = staticmethod(geom_path.draw_legend)

    @staticmethod
    def draw_group(
        data: pd.DataFrame,
        panel_params: panel_view,
        coord: Coord,
        ax: Axes,
        **params: Any,
    ):
        data.eval(
            """
            xend = x
            y = ymin
            yend = ymax
            """,
            inplace=True,
        )
        geom_segment.draw_group(data, panel_params, coord, ax, **params)



from warnings import warn

import numpy as np
from mizani.bounds import rescale_max

from ..doctools import document
from ..exceptions import PlotnineWarning
from ..utils import alias
from .scale import scale_continuous, scale_datetime, scale_discrete


@document
class scale_size_ordinal(scale_discrete):
    """
    Discrete area size scale

    Parameters
    ----------
    range : array_like
        Minimum and maximum size of the plotting symbol.
        It must be of size 2.
    {superclass_parameters}
    """

    _aesthetics = ["size"]

    def __init__(self, range=(2, 6), **kwargs):
        def palette(n):
            area = np.linspace(range[0] ** 2, range[1] ** 2, n)
            return np.sqrt(area)

        self.palette = palette
        scale_discrete.__init__(self, **kwargs)


@document
class scale_size_discrete(scale_size_ordinal):
    """
    Discrete area size scale

    Parameters
    ----------
    {superclass_parameters}
    """

    _aesthetics = ["size"]

    def __init__(self, **kwargs):
        warn(
            "Using size for a discrete variable is not advised.",
            PlotnineWarning,
        )
        super().__init__(**kwargs)


@document
class scale_size_continuous(scale_continuous):
    """
    Continuous area size scale

    Parameters
    ----------
    range : array_like
        Minimum and maximum area of the plotting symbol.
        It must be of size 2.
    {superclass_parameters}
    """

    _aesthetics = ["size"]

    def __init__(self, range=(1, 6), **kwargs):
        from mizani.palettes import area_pal

        # TODO: fix types in mizani
        self.palette = area_pal(range)  # pyright: ignore
        scale_continuous.__init__(self, **kwargs)


alias("scale_size", scale_size_continuous)


@document
class scale_size_radius(scale_continuous):
    """
    Continuous radius size scale

    Parameters
    ----------
    range : array_like
        Minimum and maximum radius of the plotting symbol.
        It must be of size 2.
    {superclass_parameters}
    """

    _aesthetics = ["size"]

    def __init__(self, range=(1, 6), **kwargs):
        from mizani.palettes import rescale_pal

        # TODO: fix types in mizani
        self.palette = rescale_pal(range)  # pyright: ignore
        scale_continuous.__init__(self, **kwargs)


@document
class scale_size_area(scale_continuous):
    """
    Continuous area size scale

    Parameters
    ----------
    max_size : float
        Maximum size of the plotting symbol.
    {superclass_parameters}
    """

    _aesthetics = ["size"]
    rescaler = staticmethod(rescale_max)

    def __init__(self, max_size=6, **kwargs):
        from mizani.palettes import abs_area

        # TODO: fix types in mizani
        self.palette = abs_area(max_size)  # pyright: ignore
        scale_continuous.__init__(self, **kwargs)


@document
class scale_size_datetime(scale_datetime):
    """
    Datetime area-size scale

    Parameters
    ----------
    range : array_like
        Minimum and maximum area of the plotting symbol.
        It must be of size 2.
    {superclass_parameters}
    """

    _aesthetics = ["size"]

    def __init__(self, range=(1, 6), **kwargs):
        from mizani.palettes import area_pal

        # TODO: fix types in mizani
        self.palette = area_pal(range)  # pyright: ignore
        scale_datetime.__init__(self, **kwargs)



from __future__ import annotations

import typing
from contextlib import suppress

from ..coords import coord_flip
from ..doctools import document
from ..exceptions import PlotnineError
from ..utils import SIZE_FACTOR, to_rgba
from .geom import geom
from .geom_path import geom_path
from .geom_polygon import geom_polygon

if typing.TYPE_CHECKING:
    from typing import Any

    import pandas as pd

    from plotnine.iapi import panel_view
    from plotnine.typing import Axes, ColorsLike, Coord


@document
class geom_ribbon(geom):
    """
    Ribbon plot

    {usage}

    Parameters
    ----------
    {common_parameters}
    outline_type : 'upper' | 'lower' | 'both' | 'full'
        How to stroke to outline of the region / area.
            * 'upper' - draw only upper bounding line'
            * 'lower' - draw only lower bounding line'
            * 'both' - draw both upper & lower bounding lines
            * 'full' - draw closed polygon around the area.
    """

    _aesthetics_doc = """
    {aesthetics_table}

    .. rubric:: Aesthetics Descriptions

    where
        Define where to exclude horizontal regions from being filled.
        Regions between any two ``False`` values are skipped.
        For sensible demarcation the value used in the *where* predicate
        expression should match the ``ymin`` value or expression. i.e.

        ::

            aes(ymin=0, ymax='col1', where='col1 > 0')  # good
            aes(ymin=0, ymax='col1', where='col1 > 10')  # bad

            aes(ymin=col2, ymax='col1', where='col1 > col2')  # good
            aes(ymin=col2, ymax='col1', where='col1 > col3')  # bad
    """
    DEFAULT_AES = {
        "alpha": 1,
        "color": "none",
        "fill": "#333333",
        "linetype": "solid",
        "size": 0.5,
        "where": True,
    }
    REQUIRED_AES = {"x", "ymax", "ymin"}
    DEFAULT_PARAMS = {
        "stat": "identity",
        "position": "identity",
        "outline_type": "both",
        "na_rm": False,
    }
    draw_legend = staticmethod(geom_polygon.draw_legend)

    def handle_na(self, data: pd.DataFrame) -> pd.DataFrame:
        return data

    def setup_data(self, data: pd.DataFrame) -> pd.DataFrame:
        # The outlines need x and y coordinates
        if self.params["outline_type"] in ("upper", "lower", "both"):
            if "xmax" in data and "x" not in data:
                data["x"] = data["xmax"]
            if "ymax" in data and "y" not in data:
                data["y"] = data["ymax"]
        return data

    @staticmethod
    def draw_group(
        data: pd.DataFrame,
        panel_params: panel_view,
        coord: Coord,
        ax: Axes,
        **params: Any,
    ):
        _x = "y" if isinstance(coord, coord_flip) else "x"
        data = coord.transform(data, panel_params, munch=True)
        data = data.sort_values(by=["group", _x], kind="mergesort")
        units = ["alpha", "color", "fill", "linetype", "size"]

        if len(data[units].drop_duplicates()) > 1:
            msg = "Aesthetics cannot vary within a ribbon."
            raise PlotnineError(msg)

        for _, udata in data.groupby(units, dropna=False):
            udata.reset_index(inplace=True, drop=True)
            geom_ribbon.draw_unit(udata, panel_params, coord, ax, **params)

    @staticmethod
    def draw_unit(
        data: pd.DataFrame,
        panel_params: panel_view,
        coord: Coord,
        ax: Axes,
        **params: Any,
    ):
        size = data["size"].iloc[0] * SIZE_FACTOR
        fill = to_rgba(data["fill"], data["alpha"])

        if data["color"].isnull().all():
            color: ColorsLike = "none"
        else:
            color = data["color"]

        if fill is None:
            fill = "none"

        if isinstance(coord, coord_flip):
            fill_between = ax.fill_betweenx
            _x, _min, _max = data["y"], data["xmin"], data["xmax"]
        else:
            fill_between = ax.fill_between
            _x, _min, _max = data["x"], data["ymin"], data["ymax"]

        # We only change this defaults for fill_between when necessary
        where = None
        interpolate = False
        with suppress(KeyError):
            if not data["where"].all():
                where = data["where"]
                interpolate = True

        if params["outline_type"] != "full":
            size = 0
            color = "none"

        fill_between(
            _x,
            _min,
            _max,  # pyright: ignore[reportGeneralTypeIssues]
            where=where,
            interpolate=interpolate,
            facecolor=fill,
            edgecolor=color,
            linewidth=size,
            linestyle=data["linetype"].iloc[0],
            zorder=params["zorder"],
            rasterized=params["raster"],
        )

        # Alpha does not affect the outlines
        data["alpha"] = 1
        geom_ribbon._draw_outline(data, panel_params, coord, ax, **params)

    @staticmethod
    def _draw_outline(
        data: pd.DataFrame,
        panel_params: panel_view,
        coord: Coord,
        ax: Axes,
        **params: Any,
    ):
        outline_type = params["outline_type"]

        if outline_type == "full":
            return

        x = "x"
        y = "y"
        if isinstance(coord, coord_flip):
            x, y = y, x
            data[x], data[y] = data[y], data[x]

        if outline_type in ("lower", "both"):
            geom_path.draw_group(
                data.eval(f"y = {y}min"), panel_params, coord, ax, **params
            )

        if outline_type in ("upper", "both"):
            geom_path.draw_group(data, panel_params, coord, ax, **params)



from plotnine import aes, geom_point, ggplot, stat_hull
from plotnine.data import mtcars


def test_hull():
    p = (
        ggplot(mtcars)
        + aes("wt", "mpg", color="factor(cyl)")
        + geom_point()
        + stat_hull(size=1)
    )

    assert p == "hull"



import numpy as np
import pandas as pd
import pandas.api.types as pdtypes

from ..doctools import document
from ..utils import resolution
from .stat import stat


@document
class stat_boxplot(stat):
    """
    Compute boxplot statistics

    {usage}

    Parameters
    ----------
    {common_parameters}
    coef : float, optional (default: 1.5)
        Length of the whiskers as a multiple of the Interquartile
        Range.

    See Also
    --------
    plotnine.geoms.geom_boxplot
    """

    _aesthetics_doc = """
    {aesthetics_table}

    .. rubric:: Options for computed aesthetics

    ::

        'width'  # width of boxplot
        'lower'  # lower hinge, 25% quantile
        'middle' # median, 50% quantile
        'upper'  # upper hinge, 75% quantile

        'notchlower' #  lower edge of notch, computed as;
                     # median - 1.58 * IQR / sqrt(n)

        'notchupper' # upper edge of notch, computed as;
                     # median + 1.58 * IQR / sqrt(n)

        'ymin'  # lower whisker, computed as; smallest observation
                # greater than or equal to lower hinge - 1.5 * IQR

        'ymax'  # upper whisker, computed as; largest observation
                # less than or equal to upper hinge + 1.5 * IQR

    Calculated aesthetics are accessed using the `after_stat` function.
    e.g. :py:`after_stat('width')`.
    """

    REQUIRED_AES = {"x", "y"}
    NON_MISSING_AES = {"weight"}
    DEFAULT_PARAMS = {
        "geom": "boxplot",
        "position": "dodge",
        "na_rm": False,
        "coef": 1.5,
        "width": None,
    }
    CREATES = {
        "lower",
        "upper",
        "middle",
        "ymin",
        "ymax",
        "outliers",
        "notchupper",
        "notchlower",
        "width",
        "relvarwidth",
    }

    def setup_params(self, data):
        if self.params["width"] is None:
            self.params["width"] = resolution(data["x"], False) * 0.75
        return self.params

    @classmethod
    def compute_group(cls, data, scales, **params):
        y = data["y"].to_numpy()
        if "weight" in data:
            weights = data["weight"]
            total_weight = np.sum(weights)
        else:
            weights = None
            total_weight = len(y)
        res = weighted_boxplot_stats(y, weights=weights, whis=params["coef"])

        if len(np.unique(data["x"])) > 1:
            width = np.ptp(data["x"]) * 0.9
        else:
            width = params["width"]

        if pdtypes.is_categorical_dtype(data["x"]):
            x = data["x"].iloc[0]
        else:
            x = np.mean([data["x"].min(), data["x"].max()])

        d = {
            "ymin": res["whislo"],
            "lower": res["q1"],
            "middle": [res["med"]],
            "upper": res["q3"],
            "ymax": res["whishi"],
            "outliers": [res["fliers"]],
            "notchupper": res["cihi"],
            "notchlower": res["cilo"],
            "x": x,
            "width": width,
            "relvarwidth": np.sqrt(total_weight),
        }
        return pd.DataFrame(d)


def weighted_percentile(a, q, weights=None):
    """
    Compute the weighted q-th percentile of data

    Parameters
    ----------
    a : array_like
        Input that can be converted into an array.
    q : array_like[float]
        Percentile or sequence of percentiles to compute. Must be int
        the range [0, 100]
    weights : array_like
        Weights associated with the input values.
    """
    # Calculate and interpolate weighted percentiles
    # method derived from https://en.wikipedia.org/wiki/Percentile
    # using numpy's standard C = 1
    if weights is None:
        weights = np.ones(len(a))

    weights = np.asarray(weights)
    q = np.asarray(q)

    C = 1
    idx_s = np.argsort(a)
    a_s = a[idx_s]
    w_n = weights[idx_s]
    S_N = np.sum(weights)
    S_n = np.cumsum(w_n)
    p_n = (S_n - C * w_n) / (S_N + (1 - 2 * C) * w_n)
    pcts = np.interp(q / 100.0, p_n, a_s)
    return pcts


def weighted_boxplot_stats(x, weights=None, whis=1.5):
    """
    Calculate weighted boxplot plot statistics

    Parameters
    ----------
    x : array_like
        Data
    weights : array_like, optional
        Weights associated with the data.
    whis : float, optional (default: 1.5)
        Position of the whiskers beyond the interquartile range.
        The data beyond the whisker are considered outliers.

        If a float, the lower whisker is at the lowest datum above
        ``Q1 - whis*(Q3-Q1)``, and the upper whisker at the highest
        datum below ``Q3 + whis*(Q3-Q1)``, where Q1 and Q3 are the
        first and third quartiles.  The default value of
        ``whis = 1.5`` corresponds to Tukey's original definition of
        boxplots.

    Notes
    -----
    This method adapted from Matplotlibs boxplot_stats. The key difference
    is the use of a weighted percentile calculation and then using linear
    interpolation to map weight percentiles back to data.
    """
    if weights is None:
        q1, med, q3 = np.percentile(x, (25, 50, 75))
        n = len(x)
    else:
        q1, med, q3 = weighted_percentile(x, (25, 50, 75), weights)
        n = np.sum(weights)

    iqr = q3 - q1
    mean = np.average(x, weights=weights)
    cilo = med - 1.58 * iqr / np.sqrt(n)
    cihi = med + 1.58 * iqr / np.sqrt(n)

    # low extreme
    loval = q1 - whis * iqr
    lox = x[x >= loval]
    if len(lox) == 0 or np.min(lox) > q1:
        whislo = q1
    else:
        whislo = np.min(lox)

    # high extreme
    hival = q3 + whis * iqr
    hix = x[x <= hival]
    if len(hix) == 0 or np.max(hix) < q3:
        whishi = q3
    else:
        whishi = np.max(hix)

    bpstats = {
        "fliers": x[(x < whislo) | (x > whishi)],
        "mean": mean,
        "med": med,
        "q1": q1,
        "q3": q3,
        "iqr": iqr,
        "whislo": whislo,
        "whishi": whishi,
        "cilo": cilo,
        "cihi": cihi,
    }
    return bpstats



from .elements import element_blank, element_line, element_rect, element_text
from .theme import theme
from .theme_gray import theme_gray


class theme_538(theme_gray):
    """
    Theme in the likeness of fivethirtyeight.com plots

    Parameters
    ----------
    base_size : int, optional
        Base font size. All text sizes are a scaled versions of
        the base font size. Default is 11.
    base_family : str, optional
        Base font family.
    """

    def __init__(self, base_size=11, base_family="DejaVu Sans"):
        super().__init__(base_size, base_family)
        bgcolor = "#F0F0F0"
        self += theme(
            axis_ticks=element_blank(),
            title=element_text(color="#3C3C3C"),
            legend_background=element_rect(fill="None"),
            legend_key=element_rect(fill="#E0E0E0"),
            panel_background=element_rect(fill=bgcolor),
            panel_border=element_blank(),
            panel_grid_major=element_line(
                color="#D5D5D5", linetype="solid", size=1
            ),
            panel_grid_minor=element_blank(),
            plot_background=element_rect(fill=bgcolor, color=bgcolor, size=1),
            strip_background=element_rect(size=0),
        )



from ..options import get_option
from .elements import element_blank, element_text
from .theme import theme


class theme_void(theme):
    """
    A classic-looking theme, with x & y axis lines and
    no gridlines.

    Parameters
    ----------
    base_size : int, optional
        Base font size. All text sizes are a scaled versions of
        the base font size. Default is 11.
    base_family : int, optional
        Base font family.
    """

    def __init__(self, base_size=11, base_family=None):
        base_family = base_family or get_option("base_family")
        m = get_option("base_margin")
        # Use only inherited elements and make everything blank
        theme.__init__(
            self,
            line=element_blank(),
            rect=element_blank(),
            text=element_text(
                family=base_family,
                style="normal",
                color="black",
                size=base_size,
                linespacing=0.9,
                ha="center",
                va="center",
                rotation=0,
                margin={},
            ),
            axis_text_x=element_blank(),
            axis_text_y=element_blank(),
            axis_title_x=element_blank(),
            axis_title_y=element_blank(),
            aspect_ratio=get_option("aspect_ratio"),
            dpi=get_option("dpi"),
            figure_size=get_option("figure_size"),
            legend_box="auto",
            legend_box_just="auto",
            legend_box_margin=0,
            legend_box_spacing=m * 3,
            legend_direction="auto",
            legend_entry_spacing_x=5,
            legend_entry_spacing_y=2,
            legend_key_height=18,
            legend_key_width=18,
            legend_margin=0,
            legend_position="right",
            legend_spacing=10,
            legend_text=element_text(
                size=base_size * 0.8,
                ha="left",
                margin={"t": 3, "b": 3, "l": 3, "r": 3, "units": "pt"},
            ),
            legend_text_legend=element_text(va="baseline"),
            legend_title=element_text(
                ha="left",
                margin={
                    "t": m,
                    "b": m,
                    "l": m,
                    "r": m,
                    "units": "fig",
                },
            ),
            legend_title_align="auto",
            panel_spacing=m,
            plot_caption=element_text(
                size=base_size * 0.8,
                ha="right",
                va="bottom",
                ma="left",
                margin={"t": m, "units": "fig"},
            ),
            plot_margin=m,
            plot_subtitle=element_text(
                size=base_size * 1,
                ha="left",
                va="top",
                ma="left",
                margin={"b": m, "units": "fig"},
            ),
            plot_title=element_text(
                size=base_size * 1.2,
                ha="left",
                va="top",
                ma="left",
                margin={"b": m, "units": "fig"},
            ),
            strip_align=0,
            strip_text=element_text(
                color="#1A1A1A",
                size=base_size * 0.8,
                linespacing=1.0,
                margin={
                    "t": 1 / 3,
                    "b": 1 / 3,
                    "l": 1 / 3,
                    "r": 1 / 3,
                    "units": "lines",
                },
            ),
            complete=True,
        )



from .elements import element_blank
from .theme import theme
from .theme_bw import theme_bw


class theme_tufte(theme_bw):
    """
    Tufte Maximal Data, Minimal Ink Theme

    Theme based on Chapter 6 'Data-Ink Maximization and Graphical
    Design of Edward Tufte *The Visual Display of Quantitative
    Information*. No border, no axis lines, no grids. This theme
    works best in combination with :class:`geom_rug()` or
    :class:`geom_rangeframe()`.

    The default font family is set to 'serif' as he uses serif
    fonts for labels in 'The Visual Display of Quantitative
    Information'. The serif font used by Tufte in his books is
    a variant of Bembo, while the sans serif font is Gill Sans.
    If these fonts are installed on your system, consider setting
    them explicitly via the argument `base_family`.

    Parameters
    ----------
    base_size : int, optional
        Base font size. All text sizes are a scaled versions of
        the base font size. Default is 11.
    base_family : str, optional
        Base font family.
    ticks: logical, optional
        Show axis ticks?

    Tufte, Edward R. (2001) The Visual Display of Quantitative
    Information, Chapter 6.

    Translated from the R ggthemes package by hyiltiz <hyiltiz@gmail.com>.
    Released under GNU GPL v2 license or later.
    """

    def __init__(self, base_size=11, base_family=None, ticks=True):
        theme_bw.__init__(self, base_size, base_family)
        d = {} if ticks else {"axis_ticks": element_blank()}
        self += theme(
            axis_line=element_blank(),
            legend_background=element_blank(),
            legend_key=element_blank(),
            panel_background=element_blank(),
            panel_border=element_blank(),
            panel_grid=element_blank(),
            plot_background=element_blank(),
            strip_background=element_blank(),
            **d,
        )



from ..doctools import document
from .stat import stat


@document
class stat_identity(stat):
    """
    Identity (do nothing) statistic

    {usage}

    Parameters
    ----------
    {common_parameters}
    """

    DEFAULT_PARAMS = {"geom": "point", "position": "identity", "na_rm": False}

    @classmethod
    def compute_panel(cls, data, scales, **params):
        return data



"""
Position Adjustments
"""
from .position_dodge import position_dodge
from .position_dodge2 import position_dodge2
from .position_fill import position_fill
from .position_identity import position_identity
from .position_jitter import position_jitter
from .position_jitterdodge import position_jitterdodge
from .position_nudge import position_nudge
from .position_stack import position_stack

__all__ = [
    "position_dodge",
    "position_dodge2",
    "position_fill",
    "position_identity",
    "position_jitter",
    "position_jitterdodge",
    "position_nudge",
    "position_stack",
]



"""
Internal API

Specifications for containers that will hold different kinds
of objects with data created when the plot is being built.
"""
from __future__ import annotations

import typing
from copy import copy
from dataclasses import dataclass, fields

if typing.TYPE_CHECKING:
    from typing import Any, Dict, Iterator, List, Optional, Sequence

    from plotnine.typing import (
        Axes,
        CoordRange,
        Figure,
        Scale,
        ScaleBreaks,
        ScaleLimits,
        StripPosition,
        TupleFloat2,
    )


@dataclass
class scale_view:
    """
    Scale information after it has been trained
    """

    scale: Scale
    aesthetics: List[str]
    name: Optional[str]
    # Trainned limits of the scale
    limits: ScaleLimits
    # Physical size of scale, including expansions
    range: CoordRange
    breaks: ScaleBreaks
    minor_breaks: Sequence[float]
    labels: Sequence[str]


@dataclass
class range_view:
    """
    Range information after trainning
    """

    range: TupleFloat2
    range_coord: TupleFloat2


@dataclass
class labels_view:
    """
    Scale labels (incl. caption & title) for the plot
    """

    x: Optional[str] = None
    y: Optional[str] = None
    alpha: Optional[str] = None
    color: Optional[str] = None
    colour: Optional[str] = None
    fill: Optional[str] = None
    linetype: Optional[str] = None
    shape: Optional[str] = None
    size: Optional[str] = None
    stroke: Optional[str] = None
    title: Optional[str] = None
    caption: Optional[str] = None
    subtitle: Optional[str] = None

    def update(self, other: labels_view):
        """
        Update the labels with those in other
        """
        for name, value in other.iter_set_fields():
            setattr(self, name, value)

    def add_defaults(self, other: labels_view):
        """
        Update labels that are missing with those in other
        """
        for name, value in other.iter_set_fields():
            cur_value = getattr(self, name)
            if cur_value is None:
                setattr(self, name, value)

    def iterfields(self) -> Iterator[tuple[str, Optional[str]]]:
        """
        Return an iterator of all (field, value) pairs
        """
        return ((f.name, getattr(self, f.name)) for f in fields(self))

    def iter_set_fields(self) -> Iterator[tuple[str, str]]:
        """
        Return an iterator of (field, value) pairs of none None values
        """
        return ((k, v) for k, v in self.iterfields() if v is not None)

    def get(self, name: str, default: str) -> str:
        """
        Get label value, return default if value is None
        """
        value = getattr(self, name)
        return str(value) if value is not None else default

    def __contains__(self, name: str) -> bool:
        """
        Return True if name has been set (is not None)
        """
        return getattr(self, name) is not None

    def __repr__(self) -> str:
        """
        Representations without the None values
        """
        nv_pairs = ", ".join(
            f"{name}={repr(value)}" for name, value in self.iter_set_fields()
        )
        return f"{self.__class__.__name__}({nv_pairs})"


@dataclass
class panel_view:
    """
    Information from the trained position scales in a panel
    """

    x: scale_view
    y: scale_view


@dataclass
class panel_ranges:
    """
    Ranges for the panel
    """

    x: TupleFloat2
    y: TupleFloat2


@dataclass
class pos_scales:
    """
    Position Scales
    """

    x: Scale
    y: Scale


@dataclass
class mpl_save_view:
    """
    Everything required to save a matplotlib figure
    """

    figure: Figure
    kwargs: Dict[str, Any]


@dataclass
class layout_details:
    """
    Layout information
    """

    panel_index: int
    panel: int
    row: int
    col: int
    scale_x: int
    scale_y: int
    axis_x: bool
    axis_y: bool
    variables: dict[str, Any]


@dataclass
class strip_draw_info:
    """
    Information required to draw strips
    """

    x: float
    y: float
    ha: str
    va: str
    box_width: float
    box_height: float
    strip_text_margin: float
    strip_align: float
    position: StripPosition
    label: str
    ax: Axes
    rotation: float


@dataclass
class strip_label_details:
    """
    Strip Label Details
    """

    # facet variable: label for the value
    variables: dict[str, str]
    meta: dict[str, Any]  # TODO: use a typeddict

    @staticmethod
    def make(
        layout_info: layout_details,
        vars: list[str],
        location: StripPosition,
    ) -> strip_label_details:
        variables: dict[str, Any] = {
            v: str(layout_info.variables[v]) for v in vars
        }
        meta: dict[str, Any] = {
            "dimension": "cols" if location == "top" else "rows"
        }
        return strip_label_details(variables, meta)

    def __len__(self) -> int:
        """
        Number of variables
        """
        return len(self.variables)

    def __copy__(self) -> strip_label_details:
        """
        Make a copy
        """
        return strip_label_details(self.variables.copy(), self.meta.copy())

    def copy(self) -> strip_label_details:
        """
        Make a copy
        """
        return copy(self)

    def text(self) -> str:
        """
        Strip text

        Join the labels for all the variables along a
        dimension
        """
        return "\n".join(list(self.variables.values()))

    def collapse(self) -> strip_label_details:
        """
        Concatenate all label values into one item
        """
        result = self.copy()
        result.variables = {"value": ", ".join(result.variables.values())}
        return result



from pathlib import Path

import pandas as pd

from plotnine import aes, geom_point, ggplot, watermark

wm_file = Path(__file__).parent / "images/plotnine-watermark.png"


def test_watermark():
    df = pd.DataFrame({"x": [1, 2, 3], "y": [1, 2, 3]})

    p = (
        ggplot(df)
        + geom_point(aes("x", "y"))
        + watermark(wm_file, 150, 160)
        + watermark(wm_file, 150, 210, 0.5)
    )

    assert p == "watermark"



from __future__ import annotations

import typing
from copy import copy

import pandas as pd
from matplotlib.animation import ArtistAnimation

from .exceptions import PlotnineError

if typing.TYPE_CHECKING:
    from typing import Iterable

    from plotnine.typing import (
        Artist,
        Axes,
        Figure,
        Ggplot,
        Scale,
    )


class PlotnineAnimation(ArtistAnimation):
    """
    Animation using ggplot objects

    Parameters
    ----------
    plots : iterable
        ggplot objects that make up the the frames of the animation
    interval : number, optional
       Delay between frames in milliseconds. Defaults to 200.
    repeat_delay : number, optional
        If the animation in repeated, adds a delay in milliseconds
        before repeating the animation. Defaults to `None`.
    repeat : bool, optional
        Controls whether the animation should repeat when the sequence
        of frames is completed. Defaults to `True`.
    blit : bool, optional
        Controls whether blitting is used to optimize drawing. Defaults
        to `False`.

    Notes
    -----
    1. The plots should have the same `facet` and
       the facet should not have fixed x and y scales.
    2. The scales of all the plots should have the same limits. It is
       a good idea to create a scale (with limits) for each aesthetic
       and add them to all the plots.
    """

    def __init__(
        self,
        plots: Iterable[Ggplot],
        interval: int = 200,
        repeat_delay: int | None = None,
        repeat: bool = True,
        blit: bool = False,
    ):
        figure, artists = self._draw_plots(plots)
        ArtistAnimation.__init__(
            self,
            figure,
            artists,
            interval=interval,
            repeat_delay=repeat_delay,
            repeat=repeat,
            blit=blit,
        )

    def _draw_plots(
        self, plots: Iterable[Ggplot]
    ) -> tuple[Figure, list[list[Artist]]]:
        with pd.option_context("mode.chained_assignment", None):
            return self.__draw_plots(plots)

    def __draw_plots(
        self, plots: Iterable[Ggplot]
    ) -> tuple[Figure, list[list[Artist]]]:
        """
        Plot and return the figure and artists

        Parameters
        ----------
        plots : iterable
            ggplot objects that make up the the frames of the animation

        Returns
        -------
        figure : matplotlib.figure.Figure
            Matplotlib figure
        artists : list
            List of :class:`Matplotlib.artist.Artist`
        """
        import matplotlib.pyplot as plt

        # For keeping track of artists for each frame
        artist_offsets: dict[str, list[int]] = {
            "collections": [],
            "patches": [],
            "lines": [],
            "texts": [],
            "artists": [],
        }

        scale_limits = {}

        def initialise_artist_offsets(n: int):
            """
            Initilise artists_offsets arrays to zero

            Parameters
            ----------
            n : int
                Number of axes to initialise artists for.
                The artists for each axes are tracked separately.
            """
            for artist_type in artist_offsets:
                artist_offsets[artist_type] = [0] * n

        def get_frame_artists(axs: list[Axes]) -> list[Artist]:
            """
            Artists shown in a given frame

            Parameters
            ----------
            axs : list[Axes]
                Matplotlib axes that have had artists added
                to them.
            """
            # The axes accumulate artists for all frames
            # For each frame we pickup the newly added artists
            # We use offsets to mark the end of the previous frame
            # e.g ax.collections[start:]
            frame_artists = []
            for i, ax in enumerate(axs):
                for name in artist_offsets:
                    start = artist_offsets[name][i]
                    new_artists = getattr(ax, name)[start:]
                    frame_artists.extend(new_artists)
                    artist_offsets[name][i] += len(new_artists)
            return frame_artists

        def set_scale_limits(scales: list[Scale]):
            """
            Set limits of all the scales in the animation

            Should be called before :func:`check_scale_limits`.

            Parameters
            ----------
            scales : list[scales]
                List of scales the have been used in building a
                ggplot object.
            """
            for sc in scales:
                ae = sc.aesthetics[0]
                scale_limits[ae] = sc.limits

        def check_scale_limits(scales: list[Scale], frame_no: int):
            """
            Check limits of the scales of a plot in the animation

            Raises a PlotnineError if any of the scales has limits
            that do not match those of the first plot/frame.

            Should be called after :func:`set_scale_limits`.

            Parameters
            ----------
            scales : list[scales]
                List of scales the have been used in building a
                ggplot object.

            frame_no : int
                Frame number
            """
            if len(scale_limits) != len(scales):
                raise PlotnineError(
                    "All plots must have the same number of scales "
                    "as the first plot of the animation."
                )

            for sc in scales:
                ae = sc.aesthetics[0]
                if ae not in scale_limits:
                    raise PlotnineError(
                        f"The plot for frame {frame_no} does not "
                        f"have a scale for the {ae} aesthetic."
                    )
                if sc.limits != scale_limits[ae]:
                    raise PlotnineError(
                        f"The {ae} scale of plot for frame {frame_no} has "
                        "different limits from those of the first frame."
                    )

        figure: Figure | None = None
        axs: list[Axes] = []
        artists = []
        scales = None  # Will hold the scales of the first frame

        # The first ggplot creates the figure, axes and the initial
        # frame of the animation. The rest of the ggplots draw
        # onto the figure and axes created by the first ggplot and
        # they create the subsequent frames.
        for frame_no, p in enumerate(plots):
            if figure is None:
                figure = p.draw()
                axs = figure.get_axes()  # pyright: ignore
                initialise_artist_offsets(len(axs))
                scales = p._build_objs.scales
                set_scale_limits(scales)
            else:
                p = copy(p)
                plot = p._draw_using_figure(figure, axs)
                check_scale_limits(plot.scales, frame_no)
            artists.append(get_frame_artists(axs))

        if figure is None:
            figure = plt.figure()  # pyright: ignore

        assert figure is not None
        # Prevent Jupyter from plotting any static figure
        plt.close(figure)
        return figure, artists



from __future__ import annotations

import typing

if typing.TYPE_CHECKING:
    import pathlib
    from typing import Any

    import matplotlib.figure

    import plotnine as p9


class watermark:
    """
    Add watermark to plot

    Parameters
    ----------
    filename : str | pathlib.Path
        Image file
    xo : int, optional
        x position offset in pixels. Default is 0.
    yo : int, optional
        y position offset in pixels. Default is 0.
    alpha : float, optional
        Alpha blending value.
    kwargs : dict
        Additional parameters passed to
        :meth:`matplotlib.figure.figimage`.

    Notes
    -----
    You can add more than one watermark to a plot.
    """

    def __init__(
        self,
        filename: str | pathlib.Path,
        xo: int = 0,
        yo: int = 0,
        alpha: float | None = None,
        **kwargs: Any,
    ):
        self.filename = filename
        kwargs.update(xo=xo, yo=yo, alpha=alpha)
        if "zorder" not in kwargs:
            kwargs["zorder"] = 99.9
        self.kwargs = kwargs

    def __radd__(self, gg: p9.ggplot) -> p9.ggplot:
        """
        Add watermark to ggplot object
        """
        gg.watermarks.append(self)
        return gg

    def draw(self, figure: matplotlib.figure.Figure):
        """
        Draw watermark

        Parameters
        ----------
        figure : Matplotlib.figure.Figure
            Matplolib figure on which to draw
        """
        from matplotlib.image import imread

        X = imread(self.filename)
        figure.figimage(X, **self.kwargs)



import numpy as np
import pandas as pd

from plotnine import (
    aes,
    coord_flip,
    geom_boxplot,
    ggplot,
    position_nudge,
)

n = 4
m = 10
df = pd.DataFrame(
    {
        "x": np.repeat([chr(65 + i) for i in range(n)], m),
        "y": (
            [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
            + [-2, 2, 3, 4, 5, 6, 7, 8, 9, 12]
            + [-7, -5, 3, 4, 5, 6, 7, 8, 12, 15]
            + [1, 2, 3, 4, 5, 8, 8, 8, 9, 10]
        ),
        "weight": (
            [4.9, 3.8, 2.7, 2.6, 1, 1, 1, 1, 1, 1]
            + [4, 3, 2, 2, 1, 1, 1, 1, 1, 1]
            + [4.9, 3.8, 2.7, 2.6, 1, 1, 1, 1, 1, 1]
            + [4, 3, 2, 2, 1, 1, 1, 1, 1, 1]
        ),
    }
)


class TestAesthetics:
    p = (
        ggplot(df, aes("x"))
        + geom_boxplot(aes(y="y"), size=2)
        + geom_boxplot(df[: 2 * m], aes(y="y+25", fill="x"), size=2)
        + geom_boxplot(df[2 * m :], aes(y="y+30", color="x"), size=2)
        + geom_boxplot(df[2 * m :], aes(y="y+55", linetype="x"), size=2)
    )

    def test_aesthetics(self):
        assert self.p == "aesthetics"

    def test_aesthetics_coordflip(self):
        assert self.p + coord_flip() == "aesthetics+coord_flip"


def test_params():
    p = (
        ggplot(df, aes("x"))
        + geom_boxplot(df[:m], aes(y="y"), size=2, notch=True)
        + geom_boxplot(
            df[m : 2 * m], aes(y="y"), size=2, notch=True, notchwidth=0.8
        )
        +
        # outliers
        geom_boxplot(
            df[2 * m : 3 * m],
            aes(y="y"),
            size=2,
            outlier_size=4,
            outlier_color="green",
        )
        + geom_boxplot(
            df[2 * m : 3 * m],
            aes(y="y+25"),
            size=2,
            outlier_size=4,
            outlier_alpha=0.5,
        )
        + geom_boxplot(
            df[2 * m : 3 * m],
            aes(y="y+60"),
            size=2,
            outlier_size=4,
            outlier_shape="D",
        )
        # position dodge
        + geom_boxplot(df[3 * m : 4 * m], aes(y="y", fill="factor(y%2)"))
    )
    assert p == "params"


def test_position_nudge():
    p = ggplot(df, aes("x", "y")) + geom_boxplot(
        position=position_nudge(x=-0.1), size=2
    )
    assert p == "position_nudge"


def test_weight():
    # The boxes of the two plots should differ slightly due to the
    # method used to calculate weighted percentiles. There is no
    # standard method for calculating weighted percentiles.
    df = pd.DataFrame(
        {
            "x": list("a" * 11 + "b" * 5),
            "y": np.hstack(
                [[1, 2, 2, 3, 3, 3, 4, 4, 4, 4, 15], [1, 2, 3, 4, 15]]
            ),
            "weight": np.hstack([np.ones(11), [1, 2, 3, 4, 1]]),
        }
    )
    p = ggplot(df, aes(x="x", y="y", weight="weight")) + geom_boxplot()
    assert p == "weight"



from __future__ import annotations

import typing

from .transforms import ZEROS_BBOX

if typing.TYPE_CHECKING:
    from matplotlib.backend_bases import RendererBase
    from matplotlib.transforms import Bbox

    from plotnine.typing import (
        Artist,
        Axes,
        Figure,
    )


def bbox_in_figure_space(
    artist: Artist, fig: Figure, renderer: RendererBase
) -> Bbox:
    """
    Bounding box of artist in figure coordinates
    """
    box = artist.get_window_extent(renderer) or ZEROS_BBOX
    return fig.transFigure.inverted().transform_bbox(box)


def tight_bbox_in_figure_space(
    artist: Artist, fig: Figure, renderer: RendererBase
) -> Bbox:
    """
    Bounding box of artist in figure coordinates
    """
    box = artist.get_tightbbox(renderer) or ZEROS_BBOX
    return fig.transFigure.inverted().transform_bbox(box)


def bbox_in_axes_space(
    artist: Artist, ax: Axes, renderer: RendererBase
) -> Bbox:
    """
    Bounding box of artist in figure coordinates
    """
    box = artist.get_window_extent(renderer) or ZEROS_BBOX
    return ax.transAxes.inverted().transform_bbox(box)


def pts_in_figure_space(fig: Figure, pts: float) -> float:
    """
    Points in figure coordinates
    """
    return fig.transFigure.inverted().transform([0, pts])[1]



from __future__ import annotations

import typing

import numpy as np
import pandas as pd

from ..exceptions import PlotnineError
from ..scales.scale import scale_discrete

if typing.TYPE_CHECKING:
    from typing import Optional

    from plotnine.typing import FloatArray, TupleFloat2


__all__ = [
    "freedman_diaconis_bins",
    "breaks_from_bins",
    "breaks_from_binwidth",
    "assign_bins",
    "fuzzybreaks",
]


def freedman_diaconis_bins(a):
    """
    Calculate number of hist bins using Freedman-Diaconis rule.
    """
    from scipy.stats import iqr

    # From http://stats.stackexchange.com/questions/798/
    a = np.asarray(a)
    h = 2 * iqr(a, nan_policy="omit") / (len(a) ** (1 / 3))

    # fall back to sqrt(a) bins if iqr is 0
    if h == 0:
        bins = np.ceil(np.sqrt(a.size))
    else:
        bins = np.ceil((np.nanmax(a) - np.nanmin(a)) / h)

    return int(bins)


def breaks_from_binwidth(
    x_range: TupleFloat2,
    binwidth: float,
    center: Optional[float] = None,
    boundary: Optional[float] = None,
):
    """
    Calculate breaks given binwidth

    Parameters
    ----------
    x_range : array_like
        Range over with to calculate the breaks. Must be
        of size 2.
    binwidth : float
        Separation between the breaks
    center : float
        The center of one of the bins
    boundary : float
        A boundary between two bins

    Returns
    -------
    out : array_like
        Sequence of break points.
    """
    if binwidth <= 0:
        raise PlotnineError("The 'binwidth' must be positive.")

    if boundary is not None and center is not None:
        raise PlotnineError(
            "Only one of 'boundary' and 'center' " "may be specified."
        )
    elif boundary is None:
        if center is None:
            # This puts min and max of data in outer half
            # of their bins
            boundary = binwidth / 2
        else:
            boundary = center - binwidth / 2

    epsilon = np.finfo(float).eps
    shift = np.floor((x_range[0] - boundary) / binwidth)
    origin = boundary + shift * binwidth
    # The (1-epsilon) factor prevents numerical roundoff in the
    # binwidth from creating an extra break beyond the one that
    # includes x_range[1].
    max_x = x_range[1] + binwidth * (1 - epsilon)
    breaks = np.arange(origin, max_x, binwidth)
    return breaks


def breaks_from_bins(x_range, bins=30, center=None, boundary=None):
    """
    Calculate breaks given binwidth

    Parameters
    ----------
    x_range : array_like
        Range over with to calculate the breaks. Must be
        of size 2.
    bins : int
        Number of bins
    center : float
        The center of one of the bins
    boundary : float
        A boundary between two bins

    Returns
    -------
    out : array_like
        Sequence of break points.
    """
    if bins < 1:
        raise PlotnineError("Need at least one bin.")
    elif bins == 1:
        binwidth = x_range[1] - x_range[0]
        boundary = x_range[1]
    else:
        binwidth = (x_range[1] - x_range[0]) / (bins - 1)

    return breaks_from_binwidth(x_range, binwidth, center, boundary)


def assign_bins(x, breaks, weight=None, pad=False, closed="right"):
    """
    Assign value in x to bins demacated by the break points

    Parameters
    ----------
    x : array_like
        Values to be binned.
    breaks : array_like
        Sequence of break points.
    weight : array_like
        Weight of each value in `x`. Used in creating the frequency
        table. If `None`, then each value in `x` has a weight of 1.
    pad : bool
        If `True`, add empty bins at either end of `x`.
    closed : str in ``['right', 'left']``
        Whether the right or left edges of the bins are part of the
        bin.

    Returns
    -------
    out : dataframe
        Bin count and density information.
    """
    right = closed == "right"
    # If weight not supplied to, use one (no weight)
    if weight is None:
        weight = np.ones(len(x))
    else:
        weight = np.asarray(weight)
        weight[np.isnan(weight)] = 0

    bin_idx = pd.cut(
        x, bins=breaks, labels=False, right=right, include_lowest=True
    )
    bin_widths = np.diff(breaks)
    bin_x = (breaks[:-1] + breaks[1:]) * 0.5

    # Create a dataframe with two columns:
    #   - the bins to which each x is assigned
    #   - the weight of each x value
    # Then create a weighted frequency table
    df = pd.DataFrame({"bin_idx": bin_idx, "weight": weight})
    wftable = df.pivot_table("weight", index=["bin_idx"], aggfunc=np.sum)[
        "weight"
    ]

    # Empty bins get no value in the computed frequency table.
    # We need to add the zeros and since frequency table is a
    # Series object, we need to keep it ordered
    if len(wftable) < len(bin_x):
        empty_bins = set(range(len(bin_x))) - set(bin_idx)
        for b in empty_bins:
            wftable.loc[b] = 0  # pyright: ignore
        wftable = wftable.sort_index()
    bin_count = wftable.tolist()

    if pad:
        bw0 = bin_widths[0]
        bwn = bin_widths[-1]
        bin_count = np.hstack([0, bin_count, 0])
        bin_widths = np.hstack([bw0, bin_widths, bwn])
        bin_x = np.hstack([bin_x[0] - bw0, bin_x, bin_x[-1] + bwn])

    return result_dataframe(bin_count, bin_x, bin_widths)


def result_dataframe(count, x, width, xmin=None, xmax=None):
    """
    Create a dataframe to hold bin information
    """
    if xmin is None:
        xmin = x - width / 2

    if xmax is None:
        xmax = x + width / 2

    # Eliminate any numerical roundoff discrepancies
    # between the edges
    xmin[1:] = xmax[:-1]
    density = (count / width) / np.sum(np.abs(count))

    out = pd.DataFrame(
        {
            "count": count,
            "x": x,
            "xmin": xmin,
            "xmax": xmax,
            "width": width,
            "density": density,
            "ncount": count / np.max(np.abs(count)),
            "ndensity": density / np.max(np.abs(density)),
            "ngroup:": np.sum(np.abs(count)),
        }
    )
    return out


def fuzzybreaks(
    scale, breaks=None, boundary=None, binwidth=None, bins=30, right=True
) -> FloatArray:
    """
    Compute fuzzy breaks

    For a continuous scale, fuzzybreaks "preserve" the range of
    the scale. The fuzzing is close to numerical roundoff and
    is visually imperceptible.

    Parameters
    ----------
    scale : scale
        Scale
    breaks : array_like
        Sequence of break points. If provided and the scale is not
        discrete, they are returned.
    boundary : float
        First break. If `None` a suitable on is computed using
        the range of the scale and the binwidth.
    binwidth : float
        Separation between the breaks
    bins : int
        Number of bins
    right : bool
        If `True` the right edges of the bins are part of the
        bin. If `False` then the left edges of the bins are part
        of the bin.

    Returns
    -------
    out : array_like
    """
    from mizani.utils import round_any

    # Bins for categorical data should take the width
    # of one level, and should show up centered over
    # their tick marks. All other parameters are ignored.
    if isinstance(scale, scale_discrete):
        breaks = scale.get_breaks()
        return -0.5 + np.arange(1, len(breaks) + 2)
    else:
        if breaks is not None:
            breaks = scale.transform(breaks)

    if breaks is not None:
        return breaks

    recompute_bins = binwidth is not None
    srange = scale.limits

    if binwidth is None or np.isnan(binwidth):
        binwidth = (srange[1] - srange[0]) / bins

    if boundary is None or np.isnan(boundary):
        boundary = round_any(srange[0], binwidth, np.floor)  # pyright: ignore

    if recompute_bins:
        bins = int(np.ceil((srange[1] - boundary) / binwidth))

    # To minimise precision errors, we do not pass the boundary and
    # binwidth into np.arange as params. The resulting breaks
    # can then be adjusted with finer(epsilon based rather than
    # some arbitrary small number) precision.
    breaks = np.arange(boundary, srange[1] + binwidth, binwidth)
    return _adjust_breaks(breaks, right)


def _adjust_breaks(breaks: FloatArray, right: bool) -> FloatArray:
    epsilon = np.finfo(float).eps
    plus = 1 + epsilon
    minus = 1 - epsilon

    sign = np.sign(breaks)
    pos_idx = np.where(sign == 1)[0]
    neg_idx = np.where(sign == -1)[0]
    zero_idx = np.where(sign == 0)[0]

    fuzzy = breaks.copy()
    if right:
        # [_](_](_](_]
        lbreak = breaks[0]
        fuzzy[pos_idx] *= plus
        fuzzy[neg_idx] *= minus
        fuzzy[zero_idx] = epsilon
        # Left closing break
        if lbreak == 0:
            fuzzy[0] = -epsilon
        elif lbreak < 0:
            fuzzy[0] = lbreak * plus
        else:
            fuzzy[0] = lbreak * minus
    else:
        # [_)[_)[_)[_]
        rbreak = breaks[-1]
        fuzzy[pos_idx] *= minus
        fuzzy[neg_idx] *= plus
        fuzzy[zero_idx] = -epsilon
        # Right closing break
        if rbreak == 0:
            fuzzy[-1] = epsilon
        elif rbreak > 0:
            fuzzy[-1] = rbreak * plus
        else:
            fuzzy[-1] = rbreak * minus

    return fuzzy



import numpy as np
import pandas as pd
import pytest

from plotnine import aes, ggplot, stat_bin, stat_density, xlim
from plotnine.exceptions import PlotnineError, PlotnineWarning


def test_stat_bin():
    x = [1, 2, 3]
    y = [1, 2, 3]
    df = pd.DataFrame({"x": x, "y": y})

    # About the default bins
    gg = ggplot(aes(x="x"), df) + stat_bin()

    with pytest.warns(PlotnineWarning) as record:
        gg.draw_test()

    res = ("bins" in str(item.message).lower() for item in record)
    assert any(res)

    # About the ignoring the y aesthetic
    gg = ggplot(aes(x="x", y="y"), df) + stat_bin()
    with pytest.raises(PlotnineError):
        gg.draw_test()


def test_changing_xlim_in_stat_density():
    n = 100
    _xlim = (5, 10)
    df = pd.DataFrame({"x": np.linspace(_xlim[0] - 1, _xlim[1] + 1, n)})
    p = ggplot(df, aes("x")) + stat_density() + xlim(*_xlim)
    # No exceptions
    with pytest.warns(PlotnineWarning):
        # warns about removed points.
        p._build()



from .elements import element_blank, element_line, element_rect, element_text
from .theme import theme, theme_get, theme_set, theme_update
from .theme_538 import theme_538
from .theme_bw import theme_bw
from .theme_classic import theme_classic
from .theme_dark import theme_dark
from .theme_gray import theme_gray, theme_grey  # type: ignore[attr-defined]
from .theme_light import theme_light
from .theme_linedraw import theme_linedraw
from .theme_matplotlib import theme_matplotlib
from .theme_minimal import theme_minimal
from .theme_seaborn import theme_seaborn
from .theme_tufte import theme_tufte
from .theme_void import theme_void
from .theme_xkcd import theme_xkcd

__all__ = [
    "element_blank",
    "element_line",
    "element_rect",
    "element_text",
    "theme",
    "theme_538",
    "theme_bw",
    "theme_classic",
    "theme_dark",
    "theme_gray",
    "theme_grey",
    "theme_light",
    "theme_linedraw",
    "theme_matplotlib",
    "theme_minimal",
    "theme_seaborn",
    "theme_void",
    "theme_xkcd",
    "theme_tufte",
    "theme_get",
    "theme_set",
    "theme_update",
]



from __future__ import annotations

import typing

import numpy as np
import pandas as pd

from ..doctools import document
from ..exceptions import PlotnineError
from ..utils import copy_missing_columns
from .geom import geom
from .geom_path import geom_path

if typing.TYPE_CHECKING:
    from typing import Any

    from plotnine.iapi import panel_view
    from plotnine.typing import Axes, Coord


@document
class geom_step(geom_path):
    """
    Stepped connected points

    {usage}

    Parameters
    ----------
    {common_parameters}
    direction : str, optional (default: hv)
        One of *hv*, *vh* or *mid*, for horizontal-vertical steps,
        vertical-horizontal steps or steps half-way between adjacent
        x values.

    See Also
    --------
    plotnine.geoms.geom_path : For documentation of extra
        parameters.
    """

    DEFAULT_PARAMS = {
        "stat": "identity",
        "position": "identity",
        "na_rm": False,
        "direction": "hv",
    }
    draw_panel = geom.draw_panel

    @staticmethod
    def draw_group(
        data: pd.DataFrame,
        panel_params: panel_view,
        coord: Coord,
        ax: Axes,
        **params: Any,
    ):
        direction = params["direction"]
        n = len(data)
        data = data.sort_values("x", kind="mergesort")
        x = data["x"].to_numpy()
        y = data["y"].to_numpy()

        if direction == "vh":
            # create stepped path -- interleave x with
            # itself and y with itself
            xidx = np.repeat(range(n), 2)[:-1]
            yidx = np.repeat(range(n), 2)[1:]
            new_x, new_y = x[xidx], y[yidx]
        elif direction == "hv":
            xidx = np.repeat(range(n), 2)[1:]
            yidx = np.repeat(range(n), 2)[:-1]
            new_x, new_y = x[xidx], y[yidx]
        elif direction == "mid":
            xidx = np.repeat(range(n - 1), 2)
            yidx = np.repeat(range(n), 2)
            diff = x[1::] - x[:-1:]
            mid_x = x[:-1:] + diff / 2
            new_x = np.hstack([x[0], mid_x[xidx], x[-1]])
            new_y = y[yidx]
        else:
            raise PlotnineError(f"Invalid direction `{direction}`")

        df = pd.DataFrame({"x": new_x, "y": new_y})
        copy_missing_columns(df, data)
        geom_path.draw_group(df, panel_params, coord, ax, **params)



from .guide_colorbar import guide_colorbar, guide_colourbar
from .guide_legend import guide_legend
from .guides import guides

__all__ = ["guide_colorbar", "guide_colourbar", "guide_legend", "guides"]



import numpy as np
import pandas as pd

from plotnine import aes, coord_flip, geom_rug, ggplot

n = 4
seq = np.arange(1, n + 1)
df = pd.DataFrame(
    {
        "x": seq,
        "y": seq,
        "z": seq,
    }
)


def test_aesthetics():
    p = (
        ggplot(df)
        + geom_rug(aes("x", "y"), size=2)
        + geom_rug(aes("x+2*n", "y+2*n", alpha="z"), size=2, sides="tr")
        + geom_rug(
            aes("x+4*n", "y+4*n", linetype="factor(z)"), size=2, sides="t"
        )
        + geom_rug(aes("x+6*n", "y+6*n", color="factor(z)"), size=2, sides="b")
        + geom_rug(aes("x+8*n", "y+8*n", size="z"), sides="tblr")
    )

    assert p == "aesthetics"


def test_coord_flip():
    p = ggplot(df) + geom_rug(aes("x", "y"), size=2, sides="l") + coord_flip()

    assert p == "coord_flip"



import matplotlib.pyplot as plt
import pytest

from plotnine import labs, lims, qplot, theme_minimal
from plotnine.animation import PlotnineAnimation
from plotnine.exceptions import PlotnineError

plt.switch_backend("Agg")  # TravisCI needs this

x = [1, 2, 3, 4, 5]
y = [1, 2, 3, 4, 5]
colors = [[1, 2, 3, 4, 5], [2, 3, 4, 5, 6], [3, 4, 5, 6, 7]]


class _PlotnineAnimation(PlotnineAnimation):
    """
    Class used for testing

    By using this class we only test the creation of
    artists and not the animation. This tests all the
    plotnine wrapper code and we can trust matplotlib
    for the rest.
    """

    def __init__(
        self, plots, interval=200, repeat_delay=None, repeat=True, blit=False
    ):
        figure, artists = self._draw_plots(plots)


def test_animation():
    def plot(i):
        return (
            qplot(x, y, color=colors[i], xlab="x", ylab="y")
            + lims(color=(1, 7))
            + labs(color="color")
            + theme_minimal()
        )

    plots = [plot(i) for i in range(3)]
    _PlotnineAnimation(plots, interval=100, repeat_delay=500)


def test_animation_different_scale_limits():
    def plot(i):
        if i == 2:
            _lims = lims(color=(3, 7))
        else:
            _lims = lims(color=(1, 7))
        return (
            qplot(x, y, color=colors[i], xlab="x", ylab="y")
            + _lims
            + labs(color="color")
            + theme_minimal()
        )

    plots = [plot(i) for i in range(3)]
    with pytest.raises(PlotnineError):
        _PlotnineAnimation(plots, interval=100, repeat_delay=500)


def test_animation_different_number_of_scales():
    def plot(i):
        if i == 2:
            p = qplot(x, y, xlab="x", ylab="y")
        else:
            p = (
                qplot(x, y, color=colors[i], xlab="x", ylab="y")
                + lims(color=(1, 7))
                + labs(color="color")
            )

        return p + theme_minimal()

    plots = [plot(i) for i in range(3)]
    with pytest.raises(PlotnineError):
        _PlotnineAnimation(plots, interval=100, repeat_delay=500)


def test_animation_different_scales():
    def plot(i):
        c = colors[i]
        if i == 2:
            p = (
                qplot(x, y, color=c, xlab="x", ylab="y")
                + lims(color=(1, 7))
                + labs(color="color")
            )
        else:
            p = (
                qplot(x, y, stroke=c, xlab="x", ylab="y")
                + lims(stroke=(1, 7))
                + labs(stroke="stroke")
            )
        return p + theme_minimal()

    plots = [plot(i) for i in range(3)]
    with pytest.raises(PlotnineError):
        _PlotnineAnimation(plots, interval=100, repeat_delay=500)



from __future__ import annotations

import re
from collections.abc import Iterable
from contextlib import suppress
from copy import deepcopy
from dataclasses import fields
from typing import Any, Dict, overload

import pandas as pd

from ..iapi import labels_view
from .evaluation import after_stat, stage

__all__ = ["aes"]

X_AESTHETICS = {"x", "xmin", "xmax", "xend", "xintercept"}
Y_AESTHETICS = {"y", "ymin", "ymax", "yend", "yintercept"}

ALL_AESTHETICS = {
    "alpha",
    "angle",
    "color",
    "colour",
    "fill",
    "group",
    "intercept",
    "label",
    "lineheight",
    "linetype",
    "lower",
    "middle",
    "radius",
    "sample",
    "shape",
    "size",
    "slope",
    "stroke",
    "upper",
    "weight",
    *X_AESTHETICS,
    *Y_AESTHETICS,
}

POSITION_AESTHETICS = X_AESTHETICS | Y_AESTHETICS

SCALED_AESTHETICS = {
    "x",
    "y",
    "alpha",
    "color",
    "colour",
    "fill",
    "linetype",
    "shape",
    "size",
    "stroke",
}

NO_GROUP = -1

# Aesthetics modifying searchers, DEPRECATED
STAT_RE = re.compile(r"\bstat\(")
DOTS_RE = re.compile(r"\.\.([a-zA-Z0-9_]+)\.\.")


class aes(Dict[str, Any]):
    """
    Create aesthetic mappings

    Parameters
    ----------
    x : expression | array_like | scalar
        x aesthetic mapping
    y : expression | array_like | scalar
        y aesthetic mapping
    **kwargs : dict
        Other aesthetic mappings

    Notes
    -----
    Only the **x** and **y** aesthetic mappings can be specified as
    positional arguments. All the rest must be keyword arguments.

    The value of each mapping must be one of:

    - **string**::

            import pandas as pd
            import numpy as np

            arr = [11, 12, 13]
            df = pd.DataFrame({'alpha': [1, 2, 3],
                               'beta': [1, 2, 3],
                               'gam ma': [1, 2, 3]})

            # Refer to a column in a dataframe
            ggplot(df, aes(x='alpha', y='beta'))

    - **array_like**::

            # A variable
            ggplot(df, aes(x='alpha', y=arr))

            # or an inplace list
            ggplot(df, aes(x='alpha', y=[4, 5, 6]))

    - **scalar**::

            # A scalar value/variable
            ggplot(df, aes(x='alpha', y=4))

            # The above statement is equivalent to
            ggplot(df, aes(x='alpha', y=[4, 4, 4]))

    - **String expression**::

            ggplot(df, aes(x='alpha', y='2*beta'))
            ggplot(df, aes(x='alpha', y='np.sin(beta)'))
            ggplot(df, aes(x='df.index', y='beta'))

            # If `count` is an aesthetic calculated by a stat
            ggplot(df, aes(x='alpha', y=after_stat('count')))
            ggplot(df, aes(x='alpha', y=after_stat('count/np.max(count)')))

      The strings in the expression can refer to;

        1. columns in the dataframe
        2. variables in the namespace
        3. aesthetic values (columns) calculated by the ``stat``

      with the column names having precedence over the variables.
      For expressions, columns in the dataframe that are mapped to
      must have names that would be valid python variable names.

      This is okay::

        # 'gam ma' is a column in the dataframe
        ggplot(df, aes(x='df.index', y='gam ma'))

      While this is not::

        # 'gam ma' is a column in the dataframe, but not
        # valid python variable name
        ggplot(df, aes(x='df.index', y='np.sin(gam ma)'))

    ``aes`` has 2 internal methods you can use to transform variables being
    mapped.

        1. ``factor`` - This function turns the variable into a factor.
            It is just an alias to ``pd.Categorical``::

                ggplot(mtcars, aes(x='factor(cyl)')) + geom_bar()

        2. ``reorder`` - This function changes the order of first variable
            based on values of the second variable::

                df = pd.DataFrame({
                    'x': ['b', 'd', 'c', 'a'],
                    'y': [1, 2, 3, 4]
                })

                ggplot(df, aes('reorder(x, y)', 'y')) + geom_col()

    .. rubric:: The group aesthetic

    ``group`` is a special aesthetic that the user can *map* to.
    It is used to group the plotted items. If not specified, it
    is automatically computed and in most cases the computed
    groups are sufficient. However, there may be cases were it is
    handy to map to it.

    See Also
    --------
    :func:`after_stat` : For how to map aesthetics to variable calculated
        by the stat
    :func:`after_scale` : For how to alter aesthetics after the data has been
        mapped by the scale.
    :class:`stage` : For how to map to evaluate the mapping to aesthetics at
        more than one stage of the plot building pipeline.
    """

    def __init__(self, *args, **kwargs):
        kwargs = rename_aesthetics(kwargs)
        kwargs.update(zip(("x", "y"), args))
        kwargs = self._convert_deprecated_expr(kwargs)
        self.update(kwargs)

    def _convert_deprecated_expr(self, kwargs):
        """
        Handle old-style calculated aesthetic expression mappings

        Just converts them to use `stage` e.g.
        'stat(count)' to after_stat(count)
        '..count..' to after_stat(count)
        """
        for name, value in kwargs.items():
            if not isinstance(value, stage):
                if is_calculated_aes(value):
                    _after_stat = strip_calculated_markers(value)
                    kwargs[name] = after_stat(_after_stat)
        return kwargs

    @property
    def _starting(self):
        """
        Return the subset of aesthetics mapped from the layer data

        The mapping is a dict of the form ``{name: expr}``, i.e the
        stage class has been peeled off.
        """
        d = {}
        for name, value in self.items():
            if not isinstance(value, stage):
                d[name] = value
            elif isinstance(value, stage) and value.start is not None:
                d[name] = value.start

        return d

    @property
    def _calculated(self):
        """
        Return only the aesthetics mapped to calculated statistics

        The mapping is a dict of the form ``{name: expr}``, i.e the
        stage class has been peeled off.
        """
        d = {}
        for name, value in self.items():
            if isinstance(value, stage) and value.after_stat is not None:
                d[name] = value.after_stat

        return d

    @property
    def _scaled(self):
        """
        Return only the aesthetics mapped to after scaling

        The mapping is a dict of the form ``{name: expr}``, i.e the
        stage class has been peeled off.
        """
        d = {}
        for name, value in self.items():
            if isinstance(value, stage) and value.after_scale is not None:
                d[name] = value.after_scale

        return d

    def __deepcopy__(self, memo):
        """
        Deep copy without copying the environment
        """
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result

        # Just copy the keys and point to the env
        for key, item in self.items():
            result[key] = deepcopy(self[key], memo)

        return result

    def __radd__(self, gg):
        """
        Add aesthetic mappings to ggplot
        """
        self = deepcopy(self)
        gg.mapping.update(self)
        gg.labels.update(make_labels(self))
        return gg

    def copy(self):
        return aes(**self)

    def inherit(self, other: dict[str, Any] | aes) -> aes:
        """
        Create a  mapping that inherits aesthetics in other

        Parameters
        ----------
        other: aes | dict[str, Any]
            Default aesthetics

        Returns
        -------
        new : aes
            Aesthetic mapping
        """
        new = self.copy()
        for k in other:
            if k not in self:
                new[k] = other[k]
        return new


@overload
def rename_aesthetics(obj: list[str]) -> list[str]:
    ...


@overload
def rename_aesthetics(obj: dict[str, Any]) -> dict[str, Any]:
    ...


def rename_aesthetics(
    obj: list[str] | dict[str, Any]
) -> list[str] | dict[str, Any]:
    """
    Rename aesthetics in obj

    Parameters
    ----------
    obj : dict or list
        Object that contains aesthetics names

    Returns
    -------
    obj : dict or list
        Object that contains aesthetics names
    """
    if isinstance(obj, dict):
        for name in tuple(obj.keys()):
            new_name = name.replace("colour", "color")
            if name != new_name:
                obj[new_name] = obj.pop(name)
    else:
        obj = [name.replace("colour", "color") for name in obj]

    return obj


def is_calculated_aes(ae):
    """
    Return True if Aesthetic expression maps to calculated statistic

    This function is now only used to identify the deprecated versions
    e.g. "..var.." or "stat(var)".

    Parameters
    ----------
    ae : object
        Single aesthetic mapping

    >>> is_calculated_aes('density')
    False

    >>> is_calculated_aes(4)
    False

    >>> is_calculated_aes('..density..')
    True

    >>> is_calculated_aes('stat(density)')
    True

    >>> is_calculated_aes('stat(100*density)')
    True

    >>> is_calculated_aes('100*stat(density)')
    True
    """
    if not isinstance(ae, str):
        return False

    for pattern in (STAT_RE, DOTS_RE):
        if pattern.search(ae):
            return True

    return False


def strip_stat(value):
    """
    Remove stat function that mark calculated aesthetics

    Parameters
    ----------
    value : object
        Aesthetic value. In most cases this will be a string
        but other types will pass through unmodified.

    Return
    ------
    out : object
        Aesthetic value with the dots removed.

    >>> strip_stat('stat(density + stat(count))')
    density + count

    >>> strip_stat('stat(density) + 5')
    density + 5

    >>> strip_stat('5 + stat(func(density))')
    5 + func(density)

    >>> strip_stat('stat(func(density) + var1)')
    func(density) + var1

    >>> strip_stat('stat + var1')
    stat + var1

    >>> strip_stat(4)
    4
    """

    def strip_hanging_closing_parens(s):
        """
        Remove leftover  parens
        """
        # Use and integer stack to track parens
        # and ignore leftover closing parens
        stack = 0
        idx = []
        for i, c in enumerate(s):
            if c == "(":
                stack += 1
            elif c == ")":
                stack -= 1
                if stack < 0:
                    idx.append(i)
                    stack = 0
                    continue
            yield c

    with suppress(TypeError):
        if STAT_RE.search(value):
            value = re.sub(r"\bstat\(", "", value)
            value = "".join(strip_hanging_closing_parens(value))

    return value


def strip_dots(value):
    """
    Remove dots(if any) that mark calculated aesthetics

    Parameters
    ----------
    value : object
        Aesthetic value. In most cases this will be a string
        but other types will pass through unmodified.

    Return
    ------
    out : object
        Aesthetic value with the dots removed.
    """
    with suppress(TypeError):
        value = DOTS_RE.sub(r"\1", value)
    return value


def strip_calculated_markers(value):
    """
    Remove markers for calculated aesthetics

    Parameters
    ----------
    value : object
        Aesthetic value. In most cases this will be a string
        but other types will pass through unmodified.

    Return
    ------
    out : object
        Aesthetic value with the dots removed.
    """
    return strip_stat(strip_dots(value))


def aes_to_scale(var):
    """
    Look up the scale that should be used for a given aesthetic
    """
    if var in {"x", "xmin", "xmax", "xend", "xintercept"}:
        var = "x"
    elif var in {"y", "ymin", "ymax", "yend", "yintercept"}:
        var = "y"
    return var


def is_position_aes(vars_):
    """
    Figure out if an aesthetic is a position aesthetic or not
    """
    try:
        return all(aes_to_scale(v) in {"x", "y"} for v in vars_)
    except TypeError:
        return aes_to_scale(vars_) in {"x", "y"}


def make_labels(mapping: dict[str, Any] | aes) -> labels_view:
    """
    Convert aesthetic mapping into text labels
    """

    def _nice_label(value: Any) -> str | None:
        if isinstance(value, pd.Series):
            return value.name  # pyright: ignore
        elif not isinstance(value, Iterable) or isinstance(value, str):
            return str(value)
        else:
            return None

    def _make_label(ae: str, value: Any) -> str | None:
        if not isinstance(value, stage):
            return _nice_label(value)
        elif value.start is None:
            if value.after_stat is not None:
                return value.after_stat
            elif value.after_scale is not None:
                return value.after_scale
            else:
                # return ''
                raise ValueError("Unknown mapping")
        else:
            if value.after_stat is not None:
                return value.after_stat
            else:
                return _nice_label(value)

    valid_names = {f.name for f in fields(labels_view)}
    return labels_view(
        **{
            str(ae): _make_label(ae, label)
            for ae, label in mapping.items()
            if ae in valid_names
        }
    )


def is_valid_aesthetic(value, ae):
    """
    Return True if `value` looks valid.

    Parameters
    ----------
    value : object
        Value to check
    ae : str
        Aesthetic name

    Returns
    -------
    out : bool
        Whether the value is of a valid looking form.

    Notes
    -----
    There are no guarantees that he value is spot on
    valid.
    """
    if ae == "linetype":
        named = {
            "solid",
            "dashed",
            "dashdot",
            "dotted",
            "_",
            "--",
            "-.",
            ":",
            "None",
            " ",
            "",
        }
        if value in named:
            return True

        # tuple of the form (offset, (on, off, on, off, ...))
        # e.g (0, (1, 2))
        conditions = [
            isinstance(value, tuple),
            isinstance(value[0], int),
            isinstance(value[1], tuple),
            len(value[1]) % 2 == 0,
            all(isinstance(x, int) for x in value[1]),
        ]
        if all(conditions):
            return True
        return False

    elif ae == "shape":
        if isinstance(value, str):
            return True

        # tuple of the form (numsides, style, angle)
        # where style is in the range [0, 3]
        # e.g (4, 1, 45)
        conditions = [
            isinstance(value, tuple),
            all(isinstance(x, int) for x in value),
            0 <= value[1] < 3,
        ]
        if all(conditions):
            return True
        return False

    elif ae in {"color", "fill"}:
        if isinstance(value, str):
            return True
        with suppress(TypeError):
            if isinstance(value, (tuple, list)) and all(
                0 <= x <= 1 for x in value
            ):
                return True
        return False

    # For any other aesthetics we return False to allow
    # for special cases to be discovered and then coded
    # for appropriately.
    return False


def has_groups(data):
    """
    Check if data is grouped

    Parameters
    ----------
    data : dataframe
        Data

    Returns
    -------
    out : bool
        If True, the data has groups.
    """
    # If any row in the group column is equal to NO_GROUP, then
    # the data all of them are and the data has no groups
    return data.loc[0, "group"] != NO_GROUP



from warnings import warn

from ..doctools import document
from ..exceptions import PlotnineError, PlotnineWarning
from ..utils import alias
from .scale import scale_continuous, scale_discrete

# All these shapes are filled
shapes = (
    "o",  # circle
    "^",  # triangle up
    "s",  # square
    "D",  # Diamond
    "v",  # triangle down
    "*",  # star
    "p",  # pentagon
    "8",  # octagon
    "<",  # triangle left
    "h",  # hexagon1
    ">",  # triangle right
    "H",  # hexagon1
    "d",  # thin diamond
)

unfilled_shapes = (
    "+",  # plus
    "x",  # x
    ".",  # point
    "1",  # tri_down
    "2",  # tri_up
    "3",  # tri_left
    "4",  # tri_right
    ",",  # pixel
    "_",  # hline
    "|",  # vline
    0,  # tickleft
    1,  # tickright
    2,  # tickup
    3,  # tickdown
    4,  # caretleft
    5,  # caretright
    6,  # caretup
    7,  # caretdown
)

# For quick lookup
FILLED_SHAPES = set(shapes)
UNFILLED_SHAPES = set(unfilled_shapes)


@document
class scale_shape(scale_discrete):
    """
    Scale for shapes

    Parameters
    ----------
    unfilled : bool
        If ``True``, then all shapes will have no interiors
        that can be a filled.
    {superclass_parameters}
    """

    _aesthetics = ["shape"]

    def __init__(self, unfilled=False, **kwargs):
        from mizani.palettes import manual_pal

        if unfilled:
            self.palette = manual_pal(unfilled_shapes)
        else:
            self.palette = manual_pal(shapes)
        scale_discrete.__init__(self, **kwargs)


@document
class scale_shape_ordinal(scale_shape):
    """
    Scale for shapes

    Parameters
    ----------
    {superclass_parameters}
    """

    _aesthetics = ["shape"]

    def __init__(self, **kwargs):
        warn(
            "Using shapes for an ordinal variable is not advised.",
            PlotnineWarning,
        )
        super().__init__(**kwargs)


class scale_shape_continuous(scale_continuous):
    """
    Continuous scale for shapes

    This is not a valid type of scale.
    """

    def __init__(self):
        raise PlotnineError("A continuous variable can not be mapped to shape")


alias("scale_shape_discrete", scale_shape)



from .elements import element_blank, element_line, element_rect
from .theme import theme
from .theme_bw import theme_bw


class theme_classic(theme_bw):
    """
    A classic-looking theme, with x & y axis lines and no gridlines

    Parameters
    ----------
    base_size : int, optional
        Base font size. All text sizes are a scaled versions of
        the base font size. Default is 11.
    base_family : str, optional
        Base font family.
    """

    def __init__(self, base_size=11, base_family=None):
        super().__init__(base_size, base_family)
        self += theme(
            panel_border=element_blank(),
            axis_line=element_line(color="black"),
            panel_grid_major=element_blank(),
            panel_grid_minor=element_blank(),
            strip_background=element_rect(colour="black", fill="None", size=1),
            legend_key=element_blank(),
        )



from __future__ import annotations

import typing
from types import SimpleNamespace

from ..iapi import panel_view
from ..positions.position import transform_position
from .coord import coord, dist_euclidean

if typing.TYPE_CHECKING:
    from typing import Optional

    import pandas as pd

    from plotnine.iapi import scale_view
    from plotnine.typing import (
        FloatArray,
        Scale,
        TupleFloat2,
    )


class coord_cartesian(coord):
    """
    Cartesian coordinate system

    Parameters
    ----------
    xlim : None | (float, float)
        Limits for x axis. If None, then they are
        automatically computed.
    ylim : None | (float, float)
        Limits for y axis. If None, then they are
        automatically computed.
    expand : bool
        If `True`, expand the coordinate axes by
        some factor. If `False`, use the limits
        from the data.
    """

    is_linear = True

    def __init__(
        self,
        xlim: Optional[TupleFloat2] = None,
        ylim: Optional[TupleFloat2] = None,
        expand: bool = True,
    ):
        self.limits = SimpleNamespace(x=xlim, y=ylim)
        self.expand = expand

    def transform(
        self, data: pd.DataFrame, panel_params: panel_view, munch: bool = False
    ) -> pd.DataFrame:
        from mizani.bounds import squish_infinite

        def squish_infinite_x(col):
            return squish_infinite(col, range=panel_params.x.range)

        def squish_infinite_y(col):
            return squish_infinite(col, range=panel_params.y.range)

        return transform_position(data, squish_infinite_x, squish_infinite_y)

    def setup_panel_params(self, scale_x: Scale, scale_y: Scale) -> panel_view:
        """
        Compute the range and break information for the panel
        """
        from mizani.transforms import identity_trans

        def get_scale_view(
            scale: Scale, coord_limits: TupleFloat2
        ) -> scale_view:
            expansion = scale.default_expansion(expand=self.expand)
            ranges = scale.expand_limits(
                scale.limits, expansion, coord_limits, identity_trans
            )
            sv = scale.view(limits=coord_limits, range=ranges.range)
            return sv

        out = panel_view(
            x=get_scale_view(scale_x, self.limits.x),
            y=get_scale_view(scale_y, self.limits.y),
        )
        return out

    def distance(
        self,
        x: pd.Series[float],
        y: pd.Series[float],
        panel_params: panel_view,
    ) -> FloatArray:
        max_dist = dist_euclidean(panel_params.x.range, panel_params.y.range)[
            0
        ]
        return dist_euclidean(x, y) / max_dist
